"""Property prediction algorithms."""

from typing import TypeVar

# from ..core import PropertyPredictor
# from ...domains.materials import ConditionPAG


S = TypeVar("S")
# T = TypeVar('T')  CLaSSVAE uses List[PropertyPredictor[S, U]] as target
U = TypeVar("U")



"""VaeRnn implementation."""

from argparse import ArgumentParser
from typing import Any, Dict, Tuple

import torch
from torch import nn

from ....arg_parser.utils import str2bool
from ....tokenizer import Tokenizer
from ..base_model import GranularEncoderDecoderModel
from ..loss import LOSS_FACTORY
from ..module import RnnDecoder, RnnEncoder
from ..utils import KLAnnealer


class VaeRnn(GranularEncoderDecoderModel):
    """VaeRnn - variational encoder using RNN with Gaussian prior and approximate posterior."""

    def __init__(
        self,
        name: str,
        position: int,
        data: Dict[str, str],
        vocab_size: int,
        embedding_size: int,
        tokenizer: Tokenizer,
        hidden_size_enc: int = 265,
        n_layers_enc: int = 2,
        hidden_size_dec: int = 265,
        n_layers_dec: int = 2,
        bidirectional: bool = False,
        latent_size: int = 196,
        teacher_forcing: bool = True,
        loss_function: str = "ce",
        kl_low: float = 0.0,
        kl_high: float = 0.1,
        kl_n_epochs: int = 100,
        kl_start_epoch: int = 0,
        inference_check_frequency: int = 50,
        **kwargs,
    ) -> None:
        """Construct VaeRnn.

        Args:
            name: model name.
            position: position of the model.
            data: data name mappings.
            vocab_size: size of the vocabulary.
            embedding_size: size of the embedding vectors.
            tokenizer: tokenizer.
            hidden_size_enc: encoder hidden size. Defaults to 256.
            n_layers_enc: number of layers for the encoder. Defaults to 2.
            hidden_size_dec: decoder hidden size. Defaults to 256.
            n_layers_dec: number of layers for the decoder. Defaults to 2.
            bidirectional: whether the RNN cell is bidirectional. Defaults to False.
            latent_size: latent size. Defaults to 196.
            teacher_forcing: whether to teacher forcing. Defaults to True.
            loss_function: loss function. Defaults to "ce".
            kl_low: low KL weight. Defaults to 0.0.
            kl_high: high KL weight. Defaults to 0.1.
            kl_n_epochs: KL number of epochs. Defaults to 100.
            kl_start_epoch: KL starting epoch. Defaults to 0.
            inference_check_frequency: frequency for checking inference quality. Defaults to 50.

        Raises:
            ValueError: in case the provided loss function is not supported.
        """
        super().__init__(name=name, data=data)
        self.position = position
        self.input_key = f"{name}_{data['input']}"
        self.target_key = f"{name}_{data['target']}"

        self.latent_size = latent_size
        self.vocab_size = vocab_size
        self.embedding_size = embedding_size
        self.teacher_forcing = teacher_forcing
        self.tokenizer = tokenizer
        self.hidden_size_enc = hidden_size_enc
        self.n_layers_enc = (n_layers_enc,)
        self.hidden_size_dec = hidden_size_dec
        self.n_layers_dec = (n_layers_dec,)
        self.hidden_factor = (2 if bidirectional else 1) * n_layers_enc

        self.loss_function_name = loss_function.lower()
        if self.loss_function_name not in LOSS_FACTORY:
            raise ValueError(
                f"loss_function={self.loss_function_name} not supported. Pick a valid one: {sorted(list(LOSS_FACTORY.keys()))}"
            )
        self.loss_function = LOSS_FACTORY[self.loss_function_name]

        self.fc_mu = nn.Linear(self.hidden_factor * hidden_size_enc, self.latent_size)
        self.fc_var = nn.Linear(self.hidden_factor * hidden_size_enc, self.latent_size)
        self.encoder = RnnEncoder(
            vocab_size=vocab_size,
            embedding_size=embedding_size,
            hidden_size=hidden_size_enc,
            n_layers=n_layers_enc,
            bidirectional=bidirectional,
        )
        self.decoder = RnnDecoder(
            vocab_size=vocab_size,
            embedding_size=embedding_size,
            hidden_size=hidden_size_dec,
            n_layers=n_layers_dec,
            latent_size=latent_size,
        )

        self.epoch_counter = 0
        self.klannealer = KLAnnealer(
            kl_low=kl_low,
            kl_high=kl_high,
            n_epochs=kl_n_epochs,
            start_epoch=kl_start_epoch,
        )
        self.inference_check_frequency = inference_check_frequency

    def decode(self, z: Any, max_len: int = 127, *args, **kwargs) -> Any:
        """Decode a latent space point.

        Args:
            z: latent point.
            max_len: maximum sequence length. Defaults to 127.

        Returns:
            tuple with decoded texts and token indices.
        """
        decoded_texts, token_indices = self.decoder.inference_direct(
            z, self.encoder.embedding, self.tokenizer, max_len=max_len
        )
        return decoded_texts, token_indices

    def _sampling_step(self, x: Any, *args, **kwargs) -> Any:
        """Run a sampling step in the model.

        Args:
            x: model input.

        Returns:
            model sampling step output.
        """
        x, input_embedding = self.encoder(x)
        mu = self.fc_mu(x)
        log_var = self.fc_var(x)
        p, q, z = self.sample(mu, log_var)
        return p, q, z, input_embedding

    def encode(self, x: Any, *args, **kwargs) -> Any:
        """Encode a sample.

        Args:
            x: input sample.

        Returns:
            latent encoding.
        """
        _, _, z, _ = self._sampling_step(x)
        return z

    def encode_decode(self, x: Any, max_len: int = 127, *args, **kwargs) -> Any:
        """Encode and decode a sample.

        Args:
            x: input sample.
            max_len: maximum sequence length. Defaults to 127.

        Returns:
            decoded sample.
        """
        z = self.encode(x)
        return self.decode(z, max_len=max_len)

    def inference(self, x: Any, *args, **kwargs) -> Any:  # type:ignore
        """Run the model in inference mode.

        Args:
            x: sample.

        Returns:
            generated output.
        """
        max_len = x.size(1)
        _, _, z, _ = self._sampling_step(x)
        return self.decode(z, max_len=max_len)

    def _run_step(self, x: Any, *args, **kwargs) -> Any:
        """Run a step in the model.

        Args:
            x: model input.

        Returns:
            model step output.
        """
        p, q, z, input_embedding = self._sampling_step(x)
        return z, self.decoder(z, input_embedding), p, q

    def step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Tuple[Any, Any, Any]:
        """Training step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        x = input_data
        x_out = target_data
        # teacher forcing
        if self.teacher_forcing:
            x_out = x_out[:, 1:].long()
            x = x[:, :-1]

        z, x_hat, p, q = self._run_step(x)

        x_hat = x_hat.view(-1, x_hat.size(-1))
        x_target = x_out.contiguous().view(-1)

        reconstruction_loss = self.loss_function(x_hat, x_target)

        log_qz = q.log_prob(z)
        log_pz = p.log_prob(z)
        kl_scaling_factor = self.klannealer(current_epoch)
        kl = log_qz - log_pz
        kl = kl.mean()
        kl_scaled = kl * kl_scaling_factor

        loss = kl_scaled + reconstruction_loss
        logs = {
            "reconstruction_loss": reconstruction_loss,
            "kl_scaled": kl_scaled,
            "kl_unscaled": kl,
            "kl_scaling_factor": kl_scaling_factor,
            "loss": loss,
        }

        return z, loss, logs

    def val_step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Any:
        """Validation step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        x = input_data
        z, loss, logs = self.step(
            input_data=input_data,
            target_data=target_data,
            device=device,
            current_epoch=current_epoch,
        )
        if current_epoch % self.inference_check_frequency == 0 and current_epoch > 0:
            decoded_texts, token_indices = self.inference(x)
            reconstructed_texts = 0
            decoded_splitted_texts = [
                text.split(self.tokenizer.eos_token, 1)[0] for text in decoded_texts
            ]
            for _, text in enumerate(decoded_splitted_texts):
                if self.tokenizer.pad_token not in text:
                    reconstructed_texts += 1
            valid_percentage = float(reconstructed_texts) / x.size(0)
            reconstructed_bits = torch.sum(
                x[:, 1:] == token_indices[:, : x[:, 1:].size(1)]
            ).item()
            reconstructed_bits_percentage = reconstructed_bits / x.numel()
            logs.update(
                {
                    "reconstructed_bits": reconstructed_bits_percentage,
                    "validity": valid_percentage,
                }
            )

        return z, loss, logs

    @staticmethod
    def add_model_specific_args(
        parent_parser: ArgumentParser, name: str, *args, **kwargs
    ) -> ArgumentParser:
        """Adding to a parser model specific arguments.

        Args:
            parent_parser: patent parser.
            name: model name.

        Returns:
            updated parser.
        """
        parser = ArgumentParser(parents=[parent_parser], add_help=False)

        parser.add_argument(f"--data_path_{name}", type=str)
        parser.add_argument(f"--data_file_{name}", type=str)
        parser.add_argument(f"--dataset_type_{name}", type=str)
        parser.add_argument(f"--position_{name}", type=int, nargs="+")
        parser.add_argument(f"--build_vocab{name}", type=str2bool)
        parser.add_argument(f"--vocab_file{name}", type=str)
        parser.add_argument(f"--input_{name}", type=str)
        parser.add_argument(f"--target_{name}", type=str)
        parser.add_argument(f"--checkpoint_path_{name}", type=str)
        parser.add_argument(f"--checkpoint_model_name_{name}", type=str)
        parser.add_argument(f"--start_from_checkpoint_{name}", type=str2bool)
        parser.add_argument(f"--freeze_weights_{name}", type=str2bool)
        parser.add_argument(f"--hidden_size_enc_{name}", type=int)
        parser.add_argument(f"--hidden_size_dec_{name}", type=int)
        parser.add_argument(f"--n_layers_enc_{name}", type=int)
        parser.add_argument(f"--n_layers_dec_{name}", type=int)
        parser.add_argument(f"--bidirectional_{name}", type=str2bool)
        parser.add_argument(f"--latent_size_{name}", type=int)
        parser.add_argument(f"--kl_low_{name}", type=float)
        parser.add_argument(f"--kl_high_{name}", type=float)
        parser.add_argument(f"--kl_n_epochs_{name}", type=int)
        parser.add_argument(f"--kl_start_epoch_{name}", type=int)
        parser.add_argument(f"--inference_check_frequency_{name}", type=int)

        return parser



"""pytorch utils for VAEs."""

import torch


def reparameterize(mu: torch.Tensor, logvar: torch.Tensor) -> torch.Tensor:
    """
    Applies reparametrization trick to obtain sample from latent space.

    Args:
        mu: the latent means of shape batch_size x latent_size.
        logvar: latent log variances, shape batch_size x latent_size.

    Returns:
        torch.Tensor: sampled Z from the latent distribution.
    """
    return torch.randn_like(mu).mul_(torch.exp(0.5 * logvar)).add_(mu)  # type:ignore



"""Model utilities."""


class KLAnnealer:
    """Annealer scaling KL weights (beta) linearly according to the number of epochs."""

    def __init__(
        self, kl_low: float, kl_high: float, n_epochs: int, start_epoch: int
    ) -> None:
        """Construct KLAnnealer.

        Args:
            kl_low: low KL weight.
            kl_high: high KL weight.
            n_epochs: number of epochs.
            start_epoch: starting epoch.
        """
        self.kl_low = kl_low
        self.kl_high = kl_high
        self.n_epochs = n_epochs
        self.start_epoch = start_epoch
        self.kl = (self.kl_high - self.kl_low) / (self.n_epochs - self.start_epoch)

    def __call__(self, epoch: int) -> float:
        """Call the annealer.

        Args:
            epoch: current epoch number.

        Returns:
            the beta weight.
        """
        k = (epoch - self.start_epoch) if epoch >= self.start_epoch else 0
        beta = self.kl_low + k * self.kl
        if beta > self.kl_high:
            beta = self.kl_high
        return beta



#!/usr/bin/env python

"""Pytorch lightning checkpoint to HF transformers converter."""


import logging
import sys
from dataclasses import dataclass, field
from typing import Optional, cast

from transformers import Trainer, TrainingArguments

from ..training_pipelines.pytorch_lightning.language_modeling.models import (
    LM_MODULE_FACTORY,
)
from .argument_parser import ArgumentParser, DataClassType


@dataclass
class PyTorchLightningToTransformersArguments:
    """PyTorchLightning to Transformers converter arguments."""

    __name__ = "pl_to_hf_converter_args"

    training_type: str = field(
        metadata={
            "help": f"Training type of the converted model, supported types: {', '.join(LM_MODULE_FACTORY.keys())}."
        },
    )
    model_name_or_path: str = field(
        metadata={"help": "Model name or path."},
    )
    ckpt: str = field(
        metadata={"help": "Path to checkpoint."},
    )
    output_path: str = field(
        metadata={"help": "Path to the converted model."},
    )
    tokenizer_name_or_path: Optional[str] = field(
        default=None,
        metadata={
            "help": "Tokenizer name or path. If not provided defaults to model_name_or_path."
        },
    )


def main() -> None:
    """Convert pytorch lightning checkpoint to HF transformers model.

    Parsing from the command line the following parameters:
        - training type of the given checkpoint.
        - model name or path, a HF's model name.
        - tokenizer name or path, a HF's tokenizer name.
        - path of the checkpoint.
        - path where the HF model will be saved.

    Raises:
        ValueError: in case the provided training type is not supported.
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    arguments = ArgumentParser(
        cast(DataClassType, PyTorchLightningToTransformersArguments)
    ).parse_args_into_dataclasses(return_remaining_strings=True)[0]

    training_type = arguments.training_type
    model_name_or_path = arguments.model_name_or_path
    tokenizer_name_or_path = arguments.tokenizer_name_or_path
    if tokenizer_name_or_path is None:
        tokenizer_name_or_path = model_name_or_path
    ckpt = arguments.ckpt
    output_path = arguments.output_path

    if training_type not in LM_MODULE_FACTORY:
        ValueError(
            f"LM training type {training_type} is not supported. Supported types: {', '.join(LM_MODULE_FACTORY.keys())}."
        )
    model_module_class = LM_MODULE_FACTORY[training_type]

    model_module = model_module_class.load_from_checkpoint(
        ckpt,
        model_args={
            "model_name_or_path": model_name_or_path,
            "tokenizer": tokenizer_name_or_path,
        },
    )

    trainer = Trainer(
        model=model_module.model,
        tokenizer=model_module.tokenizer,
        args=TrainingArguments(output_dir=output_path),
    )
    trainer.save_model()


if __name__ == "__main__":
    main()



"""Custom exception definitions."""


class S3SyncError(RuntimeError):
    """Error in syncing the cache with S3."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize S3SyncError.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "S3SyncError"
        self.title = title
        self.detail = detail
        super().__init__(detail)


class InvalidItem(ValueError):
    """Error in validating an item."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize InvalidItem.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "InvalidItem"
        self.title = title
        self.detail = detail
        super().__init__(detail)


class InvalidAlgorithmConfiguration(ValueError):
    """Error in validating an algorithm configuration."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize InvalidAlgorithmConfiguration.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "InvalidAlgorithmConfiguration"
        self.title = title
        self.detail = detail
        super().__init__(detail)


class DuplicateApplicationRegistration(ValueError):
    """Error when identifier for a registration is not unique."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize DuplicateApplicationRegistration.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "InvalidAlgorithmConfiguration"
        self.title = title
        self.detail = detail
        super().__init__(detail)


class SamplingError(TimeoutError):
    """Error when inference takes too long."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize SamplingError.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "SamplingError"
        self.title = title
        self.detail = detail
        super().__init__(detail)



"""Module configuration."""

import logging
import os
from functools import lru_cache
from typing import Optional, Set

from pydantic import BaseSettings

from .s3 import GT4SDS3Client, S3SyncError, sync_folder_with_s3

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GT4SDConfiguration(BaseSettings):
    """GT4SDConfiguration settings from environment variables.

    Default configurations for gt4sd including a read-only COS for algorithms' artifacts.
    """

    gt4sd_local_cache_path: str = os.path.join(os.path.expanduser("~"), ".gt4sd")
    gt4sd_local_cache_path_algorithms: str = "algorithms"
    gt4sd_max_number_of_stuck_calls: int = 50
    gt4sd_max_number_of_samples: int = 1000000
    gt4sd_max_runtime: int = 86400
    gt4sd_s3_host: str = "s3.mil01.cloud-object-storage.appdomain.cloud"
    gt4sd_s3_access_key: str = "a19f93a1c67949f1a31db38e58bcb7e8"
    gt4sd_s3_secret_key: str = "5748375c761a4f09c30a68cd15e218e3b27ca3e2aebd7726"
    gt4sd_s3_secure: bool = True
    gt4sd_s3_bucket: str = "algorithms"

    class Config:
        # immutable and in turn hashable, that is required for lru_cache
        frozen = True

    @staticmethod
    @lru_cache(maxsize=None)
    def get_instance() -> "GT4SDConfiguration":
        return GT4SDConfiguration()


gt4sd_configuration_instance = GT4SDConfiguration.get_instance()
logger.info(
    f"using as local cache path: {gt4sd_configuration_instance.gt4sd_local_cache_path}"
)
try:
    os.makedirs(gt4sd_configuration_instance.gt4sd_local_cache_path)
except FileExistsError:
    logger.debug("local cache path already exists")


def sync_algorithm_with_s3(prefix: Optional[str] = None) -> str:
    """Sync an algorithm in the local cache using environment variables.

    Args:
        prefix: the relative path in the bucket (both
            on S3 and locally) to match files to download. Defaults to None.

    Returns:
        str: local path using the prefix.
    """
    folder_path = os.path.join(
        gt4sd_configuration_instance.gt4sd_local_cache_path,
        gt4sd_configuration_instance.gt4sd_local_cache_path_algorithms,
    )
    try:
        sync_folder_with_s3(
            host=gt4sd_configuration_instance.gt4sd_s3_host,
            access_key=gt4sd_configuration_instance.gt4sd_s3_access_key,
            secret_key=gt4sd_configuration_instance.gt4sd_s3_secret_key,
            bucket=gt4sd_configuration_instance.gt4sd_s3_bucket,
            folder_path=folder_path,
            prefix=prefix,
            secure=gt4sd_configuration_instance.gt4sd_s3_secure,
        )
    except S3SyncError:
        logger.exception("error in syncing the cache with S3")
    return os.path.join(folder_path, prefix) if prefix is not None else folder_path


def get_cached_algorithm_path(prefix: Optional[str] = None) -> str:
    return (
        os.path.join(
            gt4sd_configuration_instance.gt4sd_local_cache_path,
            gt4sd_configuration_instance.gt4sd_local_cache_path_algorithms,
            prefix,
        )
        if prefix is not None
        else os.path.join(
            gt4sd_configuration_instance.gt4sd_local_cache_path,
            gt4sd_configuration_instance.gt4sd_local_cache_path_algorithms,
        )
    )


def get_algorithm_subdirectories_with_s3(prefix: Optional[str] = None) -> Set[str]:

    try:
        host = gt4sd_configuration_instance.gt4sd_s3_host
        access_key = gt4sd_configuration_instance.gt4sd_s3_access_key
        secret_key = gt4sd_configuration_instance.gt4sd_s3_secret_key
        secure = gt4sd_configuration_instance.gt4sd_s3_secure
        client = GT4SDS3Client(
            host=host, access_key=access_key, secret_key=secret_key, secure=secure
        )
        bucket = gt4sd_configuration_instance.gt4sd_s3_bucket
        return client.list_directories(bucket=bucket, prefix=prefix)
    except Exception:
        logger.exception("generic syncing error")
        raise S3SyncError(
            "CacheSyncingError",
            f"error in getting directories of prefix={prefix} with host={host} access_key={access_key} secret_key={secret_key} secure={secure} bucket={bucket}",
        )


def get_algorithm_subdirectories_in_cache(prefix: Optional[str] = None) -> Set[str]:
    path = get_cached_algorithm_path(prefix=prefix)
    try:
        _, dirs, _ = next(iter(os.walk(path)))
        return set(dirs)
    except StopIteration:
        return set()



#!/usr/bin/env python

"""Run training pipelines for the GT4SD."""


import logging
import sys
from dataclasses import dataclass, field
from typing import IO, Optional, Tuple, cast

from ..training_pipelines import (
    TRAINING_PIPELINE_ARGUMENTS_MAPPING,
    TRAINING_PIPELINE_MAPPING,
)
from ..training_pipelines.core import TrainingPipelineArguments
from .argument_parser import ArgumentParser, DataClass, DataClassType

logger = logging.getLogger(__name__)

SUPPORTED_TRAINING_PIPELINES = sorted(
    list(set(TRAINING_PIPELINE_ARGUMENTS_MAPPING) & set(TRAINING_PIPELINE_MAPPING))
)


@dataclass
class TrainerArguments:
    """Trainer arguments."""

    __name__ = "trainer_base_args"

    training_pipeline_name: str = field(
        metadata={
            "help": f"Training type of the converted model, supported types: {', '.join(SUPPORTED_TRAINING_PIPELINES)}."
        },
    )
    configuration_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "Configuration file for the trainining. It can be used to completely by-pass pipeline specific arguments."
        },
    )


class TrainerArgumentParser(ArgumentParser):
    """Argument parser using a custom help logic."""

    def print_help(self, file: Optional[IO[str]] = None) -> None:
        """Print help checking dynamically whether a specific pipeline is passed.

        Args:
            file: an optional I/O stream. Defaults to None, a.k.a., stdout and stderr.
        """
        try:
            help_args_set = {"-h", "--help"}
            if (
                len(set(sys.argv).union(help_args_set)) < len(help_args_set) + 2
            ):  # considering filename
                super().print_help()
                return
            args = [arg for arg in sys.argv if arg not in help_args_set]
            parsed_arguments = super().parse_args_into_dataclasses(
                args=args, return_remaining_strings=True
            )
            trainer_arguments = None
            for arguments in parsed_arguments:
                if arguments.__name__ == "trainer_base_args":
                    trainer_arguments = arguments
                    break
            if trainer_arguments:
                trainer_arguments.training_pipeline_name
                training_pipeline_arguments = TRAINING_PIPELINE_ARGUMENTS_MAPPING.get(
                    trainer_arguments.training_pipeline_name, TrainingPipelineArguments
                )
                parser = ArgumentParser(
                    tuple(
                        [TrainerArguments, *training_pipeline_arguments]  # type:ignore
                    )
                )
                parser.print_help()
        except Exception:
            super().print_help()

    def parse_json_file(self, json_file: str) -> Tuple[DataClass, ...]:
        """Overriding default .json parser.

        It by-passes all command line arguments and simply add the training pipeline.

        Args:
            json_file: JSON file containing pipeline configuration parameters.

        Returns:
            parsed arguments in a tuple of dataclasses.
        """
        number_of_dataclass_types = len(self.dataclass_types)  # type:ignore
        self.dataclass_types = [
            dataclass_type
            for dataclass_type in self.dataclass_types
            if "gt4sd.cli.trainer.TrainerArguments" not in str(dataclass_type)
        ]
        try:
            parsed_arguments = super().parse_json_file(json_file=json_file)
        except Exception:
            logger.exception(
                f"error parsing configuration file: {json_file}, printing error and exiting"
            )
            sys.exit(1)
        if number_of_dataclass_types > len(self.dataclass_types):
            self.dataclass_types.insert(0, cast(DataClassType, TrainerArguments))
        return parsed_arguments


def main() -> None:
    """
    Run a training pipeline.

    Raises:
        ValueError: in case the provided training pipeline provided is not supported.
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    base_args = TrainerArgumentParser(
        cast(DataClassType, TrainerArguments)
    ).parse_args_into_dataclasses(return_remaining_strings=True)[0]
    training_pipeline_name = base_args.training_pipeline_name
    if training_pipeline_name not in set(SUPPORTED_TRAINING_PIPELINES):
        ValueError(
            f"Training pipeline {training_pipeline_name} is not supported. Supported types: {', '.join(SUPPORTED_TRAINING_PIPELINES)}."
        )
    arguments = TRAINING_PIPELINE_ARGUMENTS_MAPPING[training_pipeline_name]
    parser = TrainerArgumentParser(tuple([TrainerArguments, *arguments]))  # type:ignore

    configuration_filepath = base_args.configuration_file
    if configuration_filepath:
        args = parser.parse_json_file(json_file=configuration_filepath)
    else:
        args = parser.parse_args_into_dataclasses(return_remaining_strings=True)
    config = {
        arg.__name__: arg.__dict__
        for arg in args
        if isinstance(arg, TrainingPipelineArguments) and isinstance(arg.__name__, str)
    }

    pipeline = TRAINING_PIPELINE_MAPPING[training_pipeline_name]
    pipeline().train(**config)


if __name__ == "__main__":
    main()



"""PaccMann\\ :superscript:`GP` Algorithm.

PaccMann\\ :superscript:`GP` generation is conditioned via gaussian processes.
"""

import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar

from typing_extensions import Protocol, runtime_checkable

from ....domains.materials import SmallMolecule, validate_molecules
from ....exceptions import InvalidItem
from ...core import AlgorithmConfiguration, GeneratorAlgorithm
from ...registry import ApplicationsRegistry
from .implementation import GPConditionalGenerator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=SmallMolecule)
Targeted = Callable[[T], Iterable[Any]]


class PaccMannGP(GeneratorAlgorithm[S, T]):
    """PaccMann\\ :superscript:`GP` Algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """Instantiate PaccMannGP ready to generate items.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for generating small molecules (SMILES) with high affinity
            for a target protein::

                configuration = PaccMannGPGenerator()
                target = {
                    "qed": {"weight": 1.0},
                    "molwt": {"target": 200},
                    "sa": {"weight": 2.0},
                    "affinity": {"protein": "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT"}
                }
                paccmann_gp = PaccMannGP(configuration=configuration, target=target)
                items = list(paccmann_gp.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to sample batches via PaccMannGP's GPConditionalGenerator.

        Args:
            configuration: helps to set up specific application of PaccMannGP.
            target: context or condition for the generation.

        Returns:
            callable with target generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: GPConditionalGenerator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate_batch

    def validate_configuration(
        self, configuration: AlgorithmConfiguration[S, T]
    ) -> AlgorithmConfiguration[S, T]:
        @runtime_checkable
        class AnyPaccMannGPConfiguration(Protocol):
            """Protocol for PaccMannGP configurations."""

            def get_conditional_generator(
                self, resources_path: str
            ) -> GPConditionalGenerator:
                ...

            def validate_item(self, item: Any) -> S:
                ...

        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AnyPaccMannGPConfiguration)
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(PaccMannGP)
class PaccMannGPGenerator(AlgorithmConfiguration[SmallMolecule, Any]):
    """
    Configuration to generate compounds controlling molecules properties.

    Implementation from the paper: https://doi.org/10.1021/acs.jcim.1c00889.
    """

    algorithm_type: ClassVar[str] = "controlled_sampling"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=32,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    temperature: float = field(
        default=1.4,
        metadata=dict(
            description="Temperature parameter for the softmax sampling in decoding."
        ),
    )
    generated_length: int = field(
        default=100,
        metadata=dict(
            description="Maximum length in tokens of the generated molcules (relates to the SMILES length)."
        ),
    )
    limit: float = field(
        default=5.0,
        metadata=dict(description="Hypercube limits in the latent space."),
    )
    acquisition_function: str = field(
        default="EI",
        metadata=dict(
            description=(
                "Acquisition function used in the Gaussian process. "
                "More details in https://scikit-optimize.github.io/stable/modules/generated/skopt.gp_minimize.html."
            )
        ),
    )
    number_of_steps: int = field(
        default=32,
        metadata=dict(description="Number of steps for an optmization round."),
    )
    number_of_initial_points: int = field(
        default=16,
        metadata=dict(description="Number of initial points evaluated."),
    )
    initial_point_generator: str = field(
        default="random",
        metadata=dict(
            description=(
                "Scheme to generate initial points. "
                "More details in https://scikit-optimize.github.io/stable/modules/generated/skopt.gp_minimize.html."
            )
        ),
    )
    seed: int = field(
        default=42,
        metadata=dict(
            description="Seed used for random number generation in the optimizer."
        ),
    )
    number_of_optimization_rounds: int = field(
        default=1,
        metadata=dict(description="Maximum number of optimization rounds."),
    )
    sampling_variance: float = field(
        default=0.1,
        metadata=dict(
            description="Variance of the Gaussian noise applied during sampling from the optimal point."
        ),
    )
    samples_for_evaluation: int = field(
        default=4,
        metadata=dict(
            description="Number of samples averaged for each minimization function evaluation."
        ),
    )
    maximum_number_of_sampling_steps: int = field(
        default=32,
        metadata=dict(
            description="Maximum number of sampling steps in an optimization round."
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Scoring functions with parameters",
            "description": "Scoring functions will be used to generate a score for the generated molecules.",
            "type": "object",
        }

    def get_conditional_generator(self, resources_path: str) -> GPConditionalGenerator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.controlled_sampling.paccmann_rl.implementation.GPConditionalGenerator.generate_batch>` method for targeted generation.
        """
        return GPConditionalGenerator(
            resources_path=resources_path,
            temperature=self.temperature,
            generated_length=self.generated_length,
            batch_size=self.batch_size,
            limit=self.limit,
            acquisition_function=self.acquisition_function,
            number_of_steps=self.number_of_steps,
            number_of_initial_points=self.number_of_initial_points,
            initial_point_generator=self.initial_point_generator,
            seed=self.seed,
            number_of_optimization_rounds=self.number_of_optimization_rounds,
            sampling_variance=self.sampling_variance,
            samples_for_evaluation=self.samples_for_evaluation,
            maximum_number_of_sampling_steps=self.maximum_number_of_sampling_steps,
        )

    def validate_item(self, item: str) -> SmallMolecule:
        """Check that item is a valid SMILES.

        Args:
            item: a generated item that is possibly not valid.

        Raises:
            InvalidItem: in case the item can not be validated.

        Returns:
            the validated SMILES.
        """
        (
            molecules,
            _,
        ) = validate_molecules([item])
        if molecules[0] is None:
            raise InvalidItem(
                title="InvalidSMILES",
                detail=f'rdkit.Chem.MolFromSmiles returned None for "{item}"',
            )
        return SmallMolecule(item)



"""GuacaMol initialization."""

from .core import (
    AaeGenerator,
    GraphGAGenerator,
    GraphMCTSGenerator,
    GuacaMolGenerator,
    MosesGenerator,
    OrganGenerator,
    SMILESGAGenerator,
    SMILESLSTMHCGenerator,
    SMILESLSTMPPOGenerator,
    VaeGenerator,
)

__all__ = [
    "GuacaMolGenerator",
    "SMILESGAGenerator",
    "GraphGAGenerator",
    "GraphMCTSGenerator",
    "SMILESLSTMHCGenerator",
    "SMILESLSTMPPOGenerator",
    "MosesGenerator",
    "VaeGenerator",
    "AaeGenerator",
    "OrganGenerator",
]



"""PyTorch Lightning training utilities."""

import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union

from pytorch_lightning import LightningDataModule, LightningModule, Trainer
from pytorch_lightning.callbacks.base import Callback
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint

from ..core import TrainingPipeline, TrainingPipelineArguments

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class PyTorchLightningTrainingPipeline(TrainingPipeline):
    """PyTorch lightining training pipelines."""

    def train(  # type: ignore
        self,
        pl_trainer_args: Dict[str, Any],
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> None:
        """Generic training function for PyTorch Lightning-based training.

        Args:
            pl_trainer_args: pytorch lightning trainer arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.
        """

        logger.info(f"Trainer arguments: {pl_trainer_args}")

        if pl_trainer_args[
            "resume_from_checkpoint"
        ] is not None and not pl_trainer_args["resume_from_checkpoint"].endswith(
            ".ckpt"
        ):
            pl_trainer_args["resume_from_checkpoint"] = None

        pl_trainer_args["callbacks"] = {
            "model_checkpoint_callback": {
                "monitor": pl_trainer_args["monitor"],
                "save_top_k": pl_trainer_args["save_top_k"],
                "mode": pl_trainer_args["mode"],
                "every_n_train_steps": pl_trainer_args["every_n_train_steps"],
                "save_last": pl_trainer_args["save_last"],
            }
        }

        del (
            pl_trainer_args["monitor"],
            pl_trainer_args["save_top_k"],
            pl_trainer_args["mode"],
            pl_trainer_args["every_n_train_steps"],
            pl_trainer_args["save_last"],
        )

        pl_trainer_args["callbacks"] = self.add_callbacks(pl_trainer_args["callbacks"])

        trainer = Trainer(**pl_trainer_args)
        data_module, model_module = self.get_data_and_model_modules(
            model_args, dataset_args
        )
        trainer.fit(model_module, data_module)

    def get_data_and_model_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> Tuple[LightningDataModule, LightningModule]:
        """Get data and model modules for training.

        Args:
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Returns:
            the data and model modules.
        """
        raise NotImplementedError(
            "Can't get data and model modules for an abstract training pipeline."
        )

    def add_callbacks(self, callback_args: Dict[str, Any]) -> List[Callback]:
        """Create the requested callbacks for training.

        Args:
            callback_args: callback arguments passed to the configuration.

        Returns:
            list of pytorch lightning callbacks.
        """

        callbacks: List[Callback] = []
        if "early_stopping_callback" in callback_args:
            callbacks.append(EarlyStopping(**callback_args["early_stopping_callback"]))

        if "model_checkpoint_callback" in callback_args:
            callbacks.append(
                ModelCheckpoint(**callback_args["model_checkpoint_callback"])
            )

        return callbacks


@dataclass
class PytorchLightningTrainingArguments(TrainingPipelineArguments):
    """
    Arguments related to pytorch lightning trainer.
    """

    __name__ = "pl_trainer_args"

    accelerator: Union[str, None] = field(
        default="ddp", metadata={"help": "Accelerator type."}
    )
    accumulate_grad_batches: int = field(
        default=1,
        metadata={
            "help": "Accumulates grads every k batches or as set up in the dict."
        },
    )
    val_check_interval: int = field(
        default=5000, metadata={"help": " How often to check the validation set."}
    )
    default_root_dir: Union[str, None] = field(
        default=None, metadata={"help": "Default path for logs and output."}
    )

    gradient_clip_val: float = field(
        default=0.0, metadata={"help": "Gradient clipping value."}
    )
    limit_val_batches: int = field(
        default=500, metadata={"help": "How much of validation dataset to check."}
    )
    log_every_n_steps: int = field(
        default=500, metadata={"help": "How often to log within steps."}
    )
    max_epochs: int = field(
        default=3,
        metadata={"help": "Stop training once this number of epochs is reached."},
    )
    resume_from_checkpoint: Union[str, None] = field(
        default=None,
        metadata={"help": "Path/URL of the checkpoint from which training is resumed."},
    )
    gpus: Union[int, None] = field(
        default=-1,
        metadata={"help": "Number of gpus to train on."},
    )
    monitor: Union[str, None] = field(
        default=None,
        metadata={"help": "Quantity to monitor in order to store a checkpoint."},
    )
    save_last: bool = field(
        default=True,
        metadata={
            "help": "When True, always saves the model at the end of the epoch to a file last.ckpt"
        },
    )
    save_top_k: Optional[int] = field(
        default=None,
        metadata={
            "help": "The best k models according to the quantity monitored will be saved."
        },
    )
    mode: str = field(
        default="min",
        metadata={"help": "Quantity to monitor in order to store a checkpoint."},
    )
    every_n_train_steps: Union[int, None] = field(
        default=None,
        metadata={"help": "Number of training steps between checkpoints."},
    )



"""Extras handling."""
# extras requirements
EXTRAS_ENABLED: bool
try:
    import AMD_Analytics  # noqa: F401
    import cog  # noqa: F401
    import pag  # noqa: F401

    EXTRAS_ENABLED = True
except ImportError:
    EXTRAS_ENABLED = False



"""Tests for AlgorithmConfiguration."""

import os
import shutil
from typing import ClassVar

import pytest

from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.configuration import GT4SDConfiguration

gt4sd_configuration_instance = GT4SDConfiguration.get_instance()


@pytest.fixture()
def development_version_path():
    # setup
    path = os.path.join(
        gt4sd_configuration_instance.gt4sd_local_cache_path,
        gt4sd_configuration_instance.gt4sd_local_cache_path_algorithms,
        "dummy",
        "algorithm",
        "config",
        "development",
    )
    os.makedirs(path, exist_ok=False)
    # test
    yield path
    # teardown
    shutil.rmtree(
        os.path.join(
            gt4sd_configuration_instance.gt4sd_local_cache_path,
            gt4sd_configuration_instance.gt4sd_local_cache_path_algorithms,
            "dummy",
        )
    )


def test_list_versions_local_only(development_version_path):
    class Config(AlgorithmConfiguration):
        algorithm_type: ClassVar[str] = "dummy"
        domain: ClassVar[str] = ""
        algorithm_name: ClassVar[str] = "algorithm"
        algorithm_application: ClassVar[str] = "config"
        algorithm_version: str = "development"

    assert "development" in Config.list_versions()



"""Graph MCTS implementation."""

import os

from guacamol_baselines.graph_mcts import goal_directed_generation
from guacamol_baselines.graph_mcts.goal_directed_generation import GB_MCTS_Generator


class GraphMCTS:
    def __init__(
        self,
        init_smiles: str,
        population_size: int,
        n_jobs: int,
        generations: int,
        patience: int,
        num_sims: float,
        max_children: int,
        max_atoms: int,
        pickle_directory: str = os.path.dirname(
            os.path.realpath(goal_directed_generation.__file__)
        ),
    ):
        """Initialize SMILESGA.

        Args:
            init_smiles: path where to load hypothesis, candidate labels and, optionally, the smiles file.
            population_size: used with n_mutations for the initial generation of smiles within the population
            n_jobs: number of concurrently running jobs
            generations: number of evolutionary generations
            patience: used for early stopping if population scores remains the same after generating molecules
            num_sims: number of times to traverse the tree,
            max_children: maximum number of childerns a node could have ,
            max_atoms: maximum number of atoms to explore to terminal the node state,
            pickle_directory: path from where to load pickle files
        """
        self.init_smiles = init_smiles
        self.pickle_directory = pickle_directory
        self.population_size = population_size
        self.max_children = max_children
        self.n_jobs = n_jobs
        self.num_sims = num_sims
        self.generations = generations
        self.patience = patience
        self.max_atoms = max_atoms

    def get_generator(self) -> GB_MCTS_Generator:
        """
        used for creating an instance of the GB_MCTS_Generator

        Returns:
            An instance of GB_MCTS_Generator
        """
        optimiser = GB_MCTS_Generator(
            pickle_directory=self.pickle_directory,
            init_smiles=self.init_smiles,
            population_size=self.population_size,
            max_children=self.max_children,
            num_sims=self.num_sims,
            generations=self.generations,
            n_jobs=self.n_jobs,
            max_atoms=self.max_atoms,
            patience=self.patience,
        )
        return optimiser



"""Types, classes, validation, etc. for the material domain."""

from typing import List, NewType, Tuple, Union

import numpy as np
import pandas as pd
from rdkit import Chem

from gt4sd.exceptions import InvalidItem

# TODO setting to str directly requires no wrapping, so wrong strings could be passed
Protein = str  # NewType('Protein', str)
SMILES = str  # NewType('SMILES', str)
SmallMolecule = SMILES
Omics = Union[np.ndarray, pd.Series]
PAG = SMILES
Molecule = Union[SmallMolecule, Protein]
Sequence = str
Property = float


def check_smiles(smiles: SMILES):
    try:
        pass  # TODO
    except Exception:
        raise InvalidItem(title="invalid SMILES", detail="Validation as SMILES failed.")


def validate_molecules(
    smiles_list: List[SMILES],
) -> Tuple[List[Chem.rdchem.Mol], List[int]]:
    """Validate molecules.

    Args:
        smiles_list: list of SMILES representing molecules.

    Returns:
        a tuple containing RDKit molecules and valid indexes.
    """
    # generate molecules from SMILES
    molecules = [
        Chem.MolFromSmiles(a_smiles, sanitize=True) for a_smiles in smiles_list
    ]
    # valid ids
    valid_ids = [
        index for index, molecule in enumerate(molecules) if molecule is not None
    ]
    return molecules, valid_ids


Bounds = Tuple[int, int]  # NewType('Bounds', Tuple[int, int])

PhotoacidityCondition = NewType("PhotoacidityCondition", Bounds)
# photoacidity_condition = PhotoacidityCondition(
#     (0, 1)
# )  # PhotoacidityCondition(Bounds((0, 1))) if Bound was a new type

ConditionPAG = Union[PhotoacidityCondition]



"""Base model for granular."""

from abc import abstractmethod
from argparse import ArgumentParser
from typing import Any, Dict, List, Tuple

import torch
from torch import nn
from torch.distributions import Distribution


class GranularBaseModel(nn.Module):
    """Base model class."""

    position: int
    from_position: List[int]

    def __init__(self, name: str, data: Dict[str, str], *args, **kwargs) -> None:
        """Construct GranularBaseModel.

        Args:
            name: model name.
            data: data name mappings.
        """
        super().__init__()
        self.name = name
        self.data = data

    def forward(self, x: Any, *args, **kwargs) -> Any:
        """Forward pass in the model.

        Args:
            x: model input.

        Returns:
            model output.
        """
        return self._run_step(x)

    @abstractmethod
    def _run_step(self, x: Any, *args, **kwargs) -> Any:
        """Run a step in the model.

        Args:
            x: model input.

        Returns:
            model step output.
        """
        pass

    @abstractmethod
    def step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Tuple[Any, Any, Any]:
        """Training step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        pass

    @abstractmethod
    def val_step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Any:
        """Validation step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        pass

    @staticmethod
    def add_model_specific_args(
        parent_parser: ArgumentParser, name: str, *args, **kwargs
    ) -> ArgumentParser:
        """Adding to a parser model specific arguments.

        Args:
            parent_parser: patent parser.
            name: model name.

        Returns:
            updated parser.
        """
        return parent_parser


class GranularEncoderDecoderModel(GranularBaseModel):
    """Autoencoder model class."""

    latent_size: int

    @abstractmethod
    def decode(self, z: Any, *args, **kwargs) -> Any:
        """Decode a latent space point.

        Args:
            z: latent point.

        Returns:
            decoded sample.
        """
        pass

    @abstractmethod
    def encode(self, x: Any, *args, **kwargs) -> Any:
        """Encode a sample.

        Args:
            x: input sample.

        Returns:
            latent encoding.
        """
        pass

    def encode_decode(self, x: Any, *args, **kwargs) -> Any:
        """Encode and decode a sample.

        Args:
            x: input sample.

        Returns:
            decoded sample.
        """
        z = self.encode(x)
        return self.decode(z)

    def inference(self, z: Any, *args, **kwargs) -> Any:
        """Run the model in inference mode.

        Args:
            z: sample.

        Returns:
            generated output.
        """
        return self.decode(z)

    def sample(
        self, mu: torch.Tensor, log_var: torch.Tensor
    ) -> Tuple[Distribution, Distribution, torch.Tensor]:
        """Sample a point from a given mean and average following a normal log-likelihood.

        Args:
            mu: mean tensor.
            log_var: log varian tensor.

        Returns:
            a tuple containing standard normal, localized normal and the sampled point.
        """
        std = torch.exp(log_var / 2.0)
        p = torch.distributions.Normal(torch.zeros_like(mu), torch.ones_like(std))
        q = torch.distributions.Normal(mu, std)
        z = q.rsample()
        return p, q, z



"""Not so domain specific code."""

# NOTE these might change a lot
# - could be abstract base classes such that other implmentations
#   can be simply registered, without explicit inheritance
# - could be simply NewType definitions if they do not implement functionality
# - organize/sort into domain specific files



"""Configuration tests."""

import os

from gt4sd.configuration import (
    GT4SDConfiguration,
    get_algorithm_subdirectories_in_cache,
    get_algorithm_subdirectories_with_s3,
)

gt4sd_configuration_instance = GT4SDConfiguration.get_instance()


def test_default_local_cache_path():
    if "GT4SD_LOCAL_CACHE_PATH" not in os.environ:
        assert os.path.dirname(
            gt4sd_configuration_instance.gt4sd_local_cache_path
        ) == os.path.expanduser("~")
        assert (
            os.path.basename(gt4sd_configuration_instance.gt4sd_local_cache_path)
            == ".gt4sd"
        )
    else:
        assert (
            gt4sd_configuration_instance.gt4sd_local_cache_path
            == os.environ["GT4SD_LOCAL_CACHE_PATH"]
        )


def test_get_algorithm_subdirectories_with_s3():
    assert isinstance(get_algorithm_subdirectories_with_s3(), set)


def test_get_algorithm_subdirectories_in_cache():
    assert isinstance(get_algorithm_subdirectories_in_cache(), set)



"""MolGX tests."""

from typing import ClassVar, Type

import pytest

from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.registry import ApplicationsRegistry
from gt4sd.extras import EXTRAS_ENABLED
from gt4sd.tests.utils import GT4SDTestSettings

if not EXTRAS_ENABLED:
    pytest.skip("Extras from custom PyPI disabled", allow_module_level=True)
else:
    from gt4sd.algorithms.generation.molgx import MolGX, MolGXQM9Generator

test_settings = GT4SDTestSettings.get_instance()


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [(MolGXQM9Generator, "generation", "materials", MolGX.__name__)],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [(MolGXQM9Generator)],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config_class",
    [(MolGXQM9Generator)],
)
def test_available_versions(config_class: Type[AlgorithmConfiguration]):
    versions = config_class.list_versions()
    assert "v0" in versions


@pytest.mark.parametrize(
    "config, algorithm",
    [
        pytest.param(
            MolGXQM9Generator,
            MolGX,
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="slow_runtime"),
        )
    ],
)
def test_generation_via_import(config, algorithm):
    algorithm = algorithm(configuration=config())
    items = list(algorithm.sample(3))
    assert len(items) == 3


@pytest.mark.parametrize(
    "algorithm_application, algorithm_type, domain, algorithm_name",
    [
        pytest.param(
            MolGXQM9Generator.__name__,
            "generation",
            "materials",
            MolGX.__name__,
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="slow_runtime"),
        ),
    ],
)
def test_generation_via_registry(
    algorithm_type, domain, algorithm_name, algorithm_application
):
    algorithm = ApplicationsRegistry.get_application_instance(
        target=None,
        algorithm_type=algorithm_type,
        domain=domain,
        algorithm_name=algorithm_name,
        algorithm_application=algorithm_application,
    )
    items = list(algorithm.sample(3))
    assert len(items) == 3



import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar

from ...core import AlgorithmConfiguration, GeneratorAlgorithm
from ...registry import ApplicationsRegistry
from .implementation import (
    AaeIterator,
    Generator,
    GraphGAIterator,
    GraphMCTSIterator,
    OrganIterator,
    SMILESGAIterator,
    SMILESLSTMHCIterator,
    SMILESLSTMPPOIterator,
    VaeIterator,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=Any)
Targeted = Callable[[T], Iterable[Any]]


class GuacaMolGenerator(GeneratorAlgorithm[S, T]):
    """GuacaMol generation algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """
        Instantiate GuacaMolGenerator ready to generate samples.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for generating molecules given a scoring function and a score::

                config = SMILESGAGenerator()
                target = {"scoring_function_name": {"target": 0.0}}
                algorithm = GuacaMolGenerator(configuration=config, target=target)
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to perform the prediction via Guacamol's generator.

        Args:
            configuration: helps to set up specific application of Guacamol.

        Returns:
            callable with target generating samples.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate_batch  # type: ignore


@ApplicationsRegistry.register_algorithm_application(GuacaMolGenerator)
class SMILESGAGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate optimizied molecules using SMILES Genetic algorithm"""

    algorithm_name: ClassVar[str] = GuacaMolGenerator.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=32,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    population_size: int = field(
        default=100,
        metadata=dict(
            description="it is used with n_mutations for the initial generation of smiles within the population"
        ),
    )
    n_mutations: int = field(
        default=200,
        metadata=dict(
            description="it is used with population size for the initial generation of smiles within the population"
        ),
    )
    n_jobs: int = field(
        default=-1,
        metadata=dict(description="number of concurrently running jobs"),
    )
    gene_size: int = field(
        default=2,
        metadata=dict(
            description="size of the gene which is used in creation of genes"
        ),
    )
    random_start: bool = field(
        default=False,
        metadata=dict(
            description="set to True to randomly choose list of SMILES for generating optimizied molecules"
        ),
    )
    generations: int = field(
        default=2,
        metadata=dict(description="number of evolutionary generations"),
    )
    patience: int = field(
        default=4,
        metadata=dict(
            description="it is used for early stopping if population scores remains the same after generating molecules"
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Scoring functions with parameters",
            "description": "Scoring functions will be used to generate a score for SMILES.",
            "type": "object",
        }

    def get_conditional_generator(self, resources_path: str) -> SMILESGAIterator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.guacamol.implementation.smiles_ga.SMILESGAIterator.generate_batch>` method for targeted generation.
        """
        return SMILESGAIterator(
            resource_path=resources_path,
            population_size=self.population_size,
            n_mutations=self.n_mutations,
            n_jobs=self.n_jobs,
            random_start=self.random_start,
            gene_size=self.gene_size,
            generations=self.generations,
            patience=self.patience,
            batch_size=self.batch_size,
        )


@ApplicationsRegistry.register_algorithm_application(GuacaMolGenerator)
class GraphGAGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate optimizied molecules using Graph-Based Genetic algorithm"""

    algorithm_name: ClassVar[str] = GuacaMolGenerator.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=1,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    population_size: int = field(
        default=100,
        metadata=dict(
            description="it is used with n_mutations for the initial generation of smiles within the population"
        ),
    )
    mutation_rate: float = field(
        default=0.01,
        metadata=dict(
            description="frequency of the new mutations in a single gene or organism over time"
        ),
    )
    offspring_size: int = field(
        default=200,
        metadata=dict(description="number of molecules to select for new population"),
    )
    n_jobs: int = field(
        default=-1,
        metadata=dict(description="number of concurrently running jobs"),
    )
    random_start: bool = field(
        default=False,
        metadata=dict(
            description="set to True to randomly choose list of SMILES for generating optimizied molecules"
        ),
    )
    generations: int = field(
        default=2,
        metadata=dict(description="number of evolutionary generations"),
    )
    patience: int = field(
        default=4,
        metadata=dict(
            description="it is used for early stopping if population scores remains the same after generating molecules"
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Scoring functions with parameters",
            "description": "Scoring functions will be used to generate a score for SMILES.",
            "type": "object",
        }

    def get_conditional_generator(self, resources_path: str) -> GraphGAIterator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.guacamol.implementation.graph_ga.GraphGAIterator.generate_batch>` method for targeted generation.
        """
        return GraphGAIterator(
            resource_path=resources_path,
            batch_size=self.batch_size,
            offspring_size=self.offspring_size,
            population_size=self.population_size,
            mutation_rate=self.mutation_rate,
            n_jobs=self.n_jobs,
            random_start=self.random_start,
            generations=self.generations,
            patience=self.patience,
        )


@ApplicationsRegistry.register_algorithm_application(GuacaMolGenerator)
class GraphMCTSGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate optimizied molecules using Graph-based Genetic Algorithm and Generative Model/Monte Carlo Tree Search for the Exploration of Chemical Space"""

    algorithm_name: ClassVar[str] = GuacaMolGenerator.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=1,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    init_smiles: str = field(
        default="",
        metadata=dict(description="initial SMILES used for generation of states."),
    )
    population_size: int = field(
        default=100,
        metadata=dict(
            description="it is used with n_mutations for the initial generation of smiles within the population"
        ),
    )
    n_jobs: int = field(
        default=-1,
        metadata=dict(description="number of concurrently running jobs"),
    )
    generations: int = field(
        default=1000,
        metadata=dict(description="number of evolutionary generations"),
    )
    patience: int = field(
        default=4,
        metadata=dict(
            description="it is used for early stopping if population scores remains the same after generating molecules"
        ),
    )
    num_sims: float = field(
        default=40,
        metadata=dict(description="number of times to traverse the tree"),
    )
    max_children: int = field(
        default=25,
        metadata=dict(description="maximum number of childerns a node could have"),
    )
    max_atoms: int = field(
        default=60,
        metadata=dict(
            description="maximum number of atoms to explore to terminal the node state"
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Scoring functions with parameters",
            "description": "Scoring functions will be used to generate a score for SMILES.",
            "type": "object",
        }

    def get_conditional_generator(self, resources_path: str) -> GraphMCTSIterator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.guacamol.implementation.graph_mcts.GraphMCTSIterator.generate_batch>` method for targeted generation.
        """
        return GraphMCTSIterator(
            init_smiles=self.init_smiles,
            batch_size=self.batch_size,
            population_size=self.population_size,
            max_children=self.max_children,
            num_sims=self.num_sims,
            generations=self.generations,
            n_jobs=self.n_jobs,
            max_atoms=self.max_atoms,
            patience=self.patience,
        )


@ApplicationsRegistry.register_algorithm_application(GuacaMolGenerator)
class SMILESLSTMHCGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate optimizied molecules using recurrent neural networks with hill climbing algorithm"""

    algorithm_name: ClassVar[str] = GuacaMolGenerator.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=1,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    n_jobs: int = field(
        default=-1,
        metadata=dict(description="number of concurrently running jobs"),
    )
    n_epochs: int = field(
        default=20,
        metadata=dict(description="number of epochs to sample"),
    )
    mols_to_sample: int = field(
        default=1024,
        metadata=dict(description="molecules sampled at each step"),
    )
    keep_top: int = field(
        default=512,
        metadata=dict(description="maximum length of a SMILES string"),
    )
    optimize_n_epochs: int = field(
        default=2,
        metadata=dict(description="number of epochs for the optimization"),
    )
    max_len: int = field(
        default=100,
        metadata=dict(description="maximum length of a SMILES string"),
    )
    optimize_batch_size: int = field(
        default=256,
        metadata=dict(description="batch size for the optimization"),
    )
    benchmark_num_samples: int = field(
        default=4096,
        metadata=dict(
            description="number of molecules to generate from final model for the benchmark"
        ),
    )
    random_start: bool = field(
        default=False,
        metadata=dict(
            description="set to True to randomly choose list of SMILES for generating optimizied molecules"
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Scoring functions with parameters",
            "description": "Scoring functions will be used to generate a score for SMILES.",
            "type": "object",
        }

    def get_conditional_generator(self, resources_path: str) -> SMILESLSTMHCIterator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.guacamol.implementation.smiles_lstm_hc.SMILESLSTMHCIterator.generate_batch>` method for targeted generation.
        """
        return SMILESLSTMHCIterator(
            resource_path=resources_path,
            batch_size=self.batch_size,
            n_epochs=self.n_epochs,
            mols_to_sample=self.mols_to_sample,
            keep_top=self.keep_top,
            optimize_n_epochs=self.optimize_n_epochs,
            max_len=self.max_len,
            optimize_batch_size=self.optimize_batch_size,
            benchmark_num_samples=self.benchmark_num_samples,
            random_start=self.random_start,
            n_jobs=self.n_jobs,
        )


@ApplicationsRegistry.register_algorithm_application(GuacaMolGenerator)
class SMILESLSTMPPOGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate optimizied molecules using recurrent neural networks with hill climbing algorithm"""

    algorithm_name: ClassVar[str] = GuacaMolGenerator.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=1,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    num_epochs: int = field(
        default=20,
        metadata=dict(description="number of epochs to sample"),
    )
    episode_size: int = field(
        default=8192,
        metadata=dict(
            description="number of molecules sampled by the policy at the start of a series of ppo updates"
        ),
    )
    optimize_batch_size: int = field(
        default=1024,
        metadata=dict(description="batch size for the optimization"),
    )
    entropy_weight: int = field(
        default=1,
        metadata=dict(description="used for calculating entropy loss"),
    )
    kl_div_weight: int = field(
        default=10,
        metadata=dict(
            description="used for calculating Kullback-Leibler divergence loss"
        ),
    )
    clip_param: float = field(
        default=0.2,
        metadata=dict(
            description="used for determining how far the new policy is from the old one"
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Scoring functions with parameters",
            "description": "Scoring functions will be used to generate a score for SMILES.",
            "type": "object",
        }

    def get_conditional_generator(self, resources_path: str) -> SMILESLSTMPPOIterator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.guacamol.implementation.smiles_lstm_ppo.SMILESLSTMPPOIterator.generate_batch>` method for targeted generation.
        """
        return SMILESLSTMPPOIterator(
            resource_path=resources_path,
            batch_size=self.batch_size,
            num_epochs=self.num_epochs,
            episode_size=self.episode_size,
            optimize_batch_size=self.optimize_batch_size,
            entropy_weight=self.entropy_weight,
            kl_div_weight=self.kl_div_weight,
            clip_param=self.clip_param,
        )


class MosesGenerator(GeneratorAlgorithm[S, T]):
    """Moses generation algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """
        Instantiate GuacaMolGenerator ready to generate samples.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for generating molecules given a scoring function and a score:

                config = AaeGenerator()
                algorithm = MosesGenerator(configuration=config, target="")
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to perform the prediction via Guacamol's generator.

        Args:
            configuration: helps to set up specific application of Guacamol.

        Returns:
            callable with target generating samples.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate_batch  # type: ignore


@ApplicationsRegistry.register_algorithm_application(MosesGenerator)
class AaeGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate molecules using Variational autoencoder"""

    algorithm_name: ClassVar[str] = MosesGenerator.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    n_samples: int = field(
        default=20,
        metadata=dict(description="Number of SMILES to generate"),
    )
    n_batch: int = field(
        default=1024,
        metadata=dict(description="Batch size for the optimization"),
    )
    max_len: int = field(
        default=100,
        metadata=dict(description="Maximum length of the generated SMILES"),
    )

    def get_conditional_generator(self, resources_path: str) -> AaeIterator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.guacamol.implementation.AaeIterator.generate_batch>` method for targeted generation.
        """
        return AaeIterator(
            resource_path=resources_path,
            n_samples=self.n_samples,
            n_batch=self.n_batch,
            max_len=self.max_len,
        )


@ApplicationsRegistry.register_algorithm_application(MosesGenerator)
class VaeGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate molecules using Adversarial autoencoder"""

    algorithm_name: ClassVar[str] = MosesGenerator.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    n_samples: int = field(
        default=20,
        metadata=dict(description="Number of SMILES to generate"),
    )
    n_batch: int = field(
        default=1024,
        metadata=dict(description="Batch size for the optimization"),
    )
    max_len: int = field(
        default=100,
        metadata=dict(description="Maximum length of the generated SMILES"),
    )

    def get_conditional_generator(self, resources_path: str) -> VaeIterator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.guacamol.implementation.VaeIterator.generate_batch>` method for targeted generation.
        """
        return VaeIterator(
            resource_path=resources_path,
            n_samples=self.n_samples,
            n_batch=self.n_batch,
            max_len=self.max_len,
        )


@ApplicationsRegistry.register_algorithm_application(MosesGenerator)
class OrganGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate molecules using Objective-Reinforced Generative Adversarial Network"""

    algorithm_name: ClassVar[str] = MosesGenerator.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    n_samples: int = field(
        default=20,
        metadata=dict(description="Number of SMILES to generate"),
    )
    n_batch: int = field(
        default=1024,
        metadata=dict(description="Batch size for the optimization"),
    )
    max_len: int = field(
        default=100,
        metadata=dict(description="Maximum length of the generated SMILES"),
    )

    def get_conditional_generator(self, resources_path: str) -> OrganIterator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.guacamol.implementation.OrganIterator.generate_batch>` method for targeted generation.
        """
        return OrganIterator(
            resource_path=resources_path,
            n_samples=self.n_samples,
            n_batch=self.n_batch,
            max_len=self.max_len,
        )



"""enzeptional - data processing utilities."""

from abc import ABC
from typing import Generic, List, Optional, Tuple, TypeVar, Union

import numpy as np
import torch
from tape.datasets import pad_sequences
from tape.registry import registry
from tape.tokenizers import TAPETokenizer
from transformers import AutoModelWithLMHead, AutoTokenizer

from ..torch import device_claim

T = TypeVar("T")  # used for sample embedding


class Embedding(ABC, Generic[T]):
    """Abstract embedding class."""

    def embed_one(self, sample: T) -> np.ndarray:
        """Embed one sample.

        Args:
            sample: sample representation.

        Returns:
            embedding vector for the sample.
        """
        return self.__call__([sample])

    def __call__(self, samples: List[T]) -> np.ndarray:
        """Embed multiple samples sample.

        Args:
            samples: a list of sample representations.

        Returns:
            embedding vectors for the samples.
        """
        raise NotImplementedError


StringEmbedding = Embedding[str]


class TAPEEmbedding(StringEmbedding):
    """Embed AA sequence using TAPE."""

    def __init__(
        self,
        model_type: str = "transformer",
        model_dir: str = "bert-base",
        aa_vocabulary: str = "iupac",
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """Initialize the TAPE embedding class.

        Args:
            model_type: TAPE model type. Defaults to "transformer".
            model_dir: model directory. Defaults to "bert-base".
            aa_vocabulary: type of vocabulary. Defaults to "iupac".
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        # get device
        self.device = device_claim(device)
        # task and model definition
        self.task_specification = registry.get_task_spec("embed")
        self.model = registry.get_task_model(
            model_type, self.task_specification.name, load_dir=model_dir
        )
        self.model = self.model.to(self.device)
        self.model.eval()
        self.tokenizer = TAPETokenizer(vocab=aa_vocabulary)

    def _encode_and_mask(self, sequence: str) -> Tuple[np.ndarray, np.ndarray]:
        """Encode and mask a sequence.

        Args:
            sequence: AA sequence.

        Returns:
            a tuple containing the token ids and the mask.
        """
        token_ids = self.tokenizer.encode(sequence)
        return token_ids, np.ones_like(token_ids)

    def __call__(self, samples: List[str]) -> np.ndarray:
        """Embed multiple protein sequences using TAPE.

        Args:
            samples: a list of protein sequences.

        Returns:
            a numpy array containing the embedding vectors.
        """
        # prepare input
        token_ids, masks = zip(
            *[self._encode_and_mask(sequence) for sequence in samples]
        )
        input_data = {
            "input_ids": torch.from_numpy(pad_sequences(token_ids)).to(self.device),
            "input_mask": torch.from_numpy(pad_sequences(masks)).to(self.device),
        }
        sequence_lenghts = input_data["input_mask"].sum(1)
        sequence_embeddings = self.model(**input_data)[0].cpu().detach().numpy()
        # get average embedding
        return np.array(
            [
                sequence_embedding[:sequence_length].mean(0)
                for sequence_embedding, sequence_length in zip(  # type:ignore
                    sequence_embeddings, sequence_lenghts
                )
            ]
        )


class HuggingFaceTransformerEmbedding(StringEmbedding):
    """Embed a string representation of a molecule using an HF transformers model."""

    def __init__(
        self,
        model_name: str = "seyonec/ChemBERTa-zinc-base-v1",
        tokenizer_name: str = "seyonec/ChemBERTa-zinc-base-v1",
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """Initialize the HF transformers embedding class.

        Args:
            model_name: model name. Defaults to "seyonec/ChemBERTa-zinc-base-v1".
            tokenizer_name: tokenizer name. Defaults to "seyonec/ChemBERTa-zinc-base-v1".
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        # get device
        self.device = device_claim(device)
        # tokenizer and model definition
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModelWithLMHead.from_pretrained(tokenizer_name)
        self.model = self.model.to(self.device)
        self.model.eval()

    def __call__(self, samples: List[str]) -> np.ndarray:
        """Embed multiple protein sequences using TAPE.

        Args:
            samples: a list of strings representing molecules.

        Returns:
            a numpy array containing the embedding vectors.
        """
        # get the CLS token representation from each SMILES.
        return (
            self.model(
                **{
                    key: tensor.to(self.device)
                    for key, tensor in self.tokenizer(
                        samples, return_tensors="pt", padding=True
                    ).items()
                }
            )[0][:, 0, :]
            .detach()
            .numpy()
        )


def mutate_sequence_with_variant(sequence: str, variant: str) -> str:
    """Given an AA sequence and a variant returns the mutated AA sequence.

    Args:
        sequence: an AA sequence.
        variant: a variant annotation.

    Returns:
        the mutated sequence.
    """
    edits = [
        (int(variant_string[1:-1]), variant[0], variant_string[-1])
        for variant_string in map(str.strip, variant.split("/"))
    ]
    mutated_sequence = list(sequence)
    for index, _, aa_to in edits:
        mutated_sequence[index] = aa_to
    return "".join(mutated_sequence)


def sanitize_intervals(intervals: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
    """Sanitize intervals merging overlapping ones and sorting them.

    Args:
        intervals: intervals to sanitize.

    Returns:
        sorted and non overlapping intervals.
    """
    sorted_intervals = sorted(intervals, key=lambda interval: interval[0])
    merged_intervals = [sorted_intervals[0]]
    for current in sorted_intervals:
        previous_end = merged_intervals[-1][1]
        if current[0] <= previous_end:
            previous_end = max(previous_end, current[1])
        else:
            merged_intervals.append(current)
    return merged_intervals


def reconstruct_sequence_with_mutation_range(
    sequence: str, mutated_sequence_range: str, intervals: List[Tuple[int, int]]
):
    """Reconstruct a sequence replacing in given positions sub-sequences from a mutated range.

    Args:
        sequence: original sequence.
        mutated_sequence_range: mutated sequence range.
        intervals: sorted and non overlapping intervals.

    Returns:
        reconstructed sequence.
    """
    # create the mutated sequence, considering sorted intervals
    mutated_range_offset = 0  # offset with respect to the mutated_sequence_range
    mutated_sequence_offset = 0  # offset with respect to the full mutated sequence.
    mutated_sequence = ""
    for start, end in intervals:
        mutated_sequence += sequence[mutated_sequence_offset:start]
        chunk_length = end - start + 1
        mutated_sequence += mutated_sequence_range[
            mutated_range_offset : mutated_range_offset + chunk_length
        ]
        mutated_range_offset += chunk_length
        mutated_sequence_offset = end + 1
    mutated_sequence += sequence[end + 1 :]
    return mutated_sequence



"""Algortihms for topic modelling using zero-shot learning via MLNI models."""

import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar

from ...core import AlgorithmConfiguration, GeneratorAlgorithm
from ...registry import ApplicationsRegistry
from .implementation import ZeroShotClassifier

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=Any)
Targeted = Callable[[T], Iterable[Any]]


class TopicsZeroShot(GeneratorAlgorithm[S, T]):
    """Topics prediction algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """Instantiate TopicsZeroShot ready to predict topics.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for predicting topics for a given text::

                config = TopicsPredictor()
                algorithm = TopicsZeroShot(configuration=config, target="This is a text I want to understand better")
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to perform the prediction via TopicsZeroShot's generator.

        Args:
            configuration: helps to set up specific application of TopicsZeroShot.
            target: context or condition for the generation.

        Returns:
            callable with target predicting topics sorted by relevance.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: ZeroShotClassifier = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.predict


@ApplicationsRegistry.register_algorithm_application(TopicsZeroShot)
class TopicsPredictor(AlgorithmConfiguration[str, str]):
    """Configuration to generate topics."""

    algorithm_type: ClassVar[str] = "prediction"
    domain: ClassVar[str] = "nlp"
    algorithm_version: str = "dbpedia"

    model_name: str = field(
        default="facebook/bart-large-mnli",
        metadata=dict(
            description="MLNI model name to use. If the  model is not found in the cache, a download from HuggingFace will be attempted."
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Text to analyze",
            "description": "Text considered for the topics prediction task.",
            "type": "string",
        }

    def get_conditional_generator(self, resources_path: str) -> ZeroShotClassifier:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.prediction.topics_zero_shot.implementation.ZeroShotClassifier.predict>` method for targeted generation.
        """
        return ZeroShotClassifier(
            resources_path=resources_path, model_name=self.model_name
        )



"""Advanced manufacturing algorithms."""

import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar

from ....domains.materials import SmallMolecule
from ...core import AlgorithmConfiguration, GeneratorAlgorithm
from ...registry import ApplicationsRegistry
from .implementation.core import Generator
from .implementation.nccr import CatalystGenerator as NCCRCatalystGenerator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=Any)
Targeted = Callable[[T], Iterable[Any]]


class AdvancedManufacturing(GeneratorAlgorithm[S, T]):
    """Advance manufacturing generator algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """Instantiate AdvancedManufacturing ready to generate items.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for generating small molecules (SMILES) with a target binding energy::

                config = CatalystGenerator()
                algorithm = AdvancedManufacturing(configuration=config, target=10.0)
                items = list(algorithm.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to sample batches via AdvancedManufacturing's generator.

        Args:
            configuration: helps to set up specific application of AdvancedManufacturing.
            target: context or condition for the generation.

        Returns:
            callable with target generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate_samples


@ApplicationsRegistry.register_algorithm_application(AdvancedManufacturing)
class CatalystGenerator(AlgorithmConfiguration[SmallMolecule, float]):
    """Configuration to generate catalysts with a desired binding energy."""

    algorithm_type: ClassVar[str] = "controlled_sampling"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    number_of_points: int = field(
        default=32,
        metadata=dict(
            description="Number of points to sample with the Gaussian Process."
        ),
    )
    number_of_steps: int = field(
        default=50,
        metadata=dict(
            description="Number of optimization steps in the Gaussian Process optimization."
        ),
    )
    generated_length: int = field(
        default=100,
        metadata=dict(
            description="Maximum length in tokens of the generated molcules (relates to the SMILES length)."
        ),
    )
    primer_smiles: str = field(
        default="",
        metadata=dict(
            description="Primer molecule to initiate the sampling in SMILES format. Defaults to no primer."
        ),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Target energy",
            "description": "Binding energy target for the catalysts generated.",
            "type": "number",
        }

    def get_conditional_generator(self, resources_path: str) -> Generator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.controlled_sampling.advanced_manufacturing.implementation.core.Generator.generate_samples>` method for targeted generation.
        """
        return NCCRCatalystGenerator(
            resources_path=resources_path,
            generated_length=self.generated_length,
            number_of_points=self.number_of_points,
            number_of_steps=self.number_of_steps,
            primer_smiles=self.primer_smiles,
        )



"""Language modeling trainer unit tests."""

from typing import cast

import pkg_resources
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint

from gt4sd.training_pipelines import (
    TRAINING_PIPELINE_MAPPING,
    LanguageModelingTrainingPipeline,
)
from gt4sd.training_pipelines.pytorch_lightning.language_modeling.lm_datasets import (
    CGMDataModule,
    CLMDataModule,
    MLMDataModule,
    PLMDataModule,
)
from gt4sd.training_pipelines.pytorch_lightning.language_modeling.models import (
    CGMModule,
    CLMModule,
    MLMModule,
    PLMModule,
)

template_config = {
    "model_args": {
        "tokenizer": "albert-base-v2",
        "model_name_or_path": "albert-base-v2",
        "model_config_name": "albert-base-v2",
        "type": "mlm",
        "lr": 2e-5,
        "lr_decay": 0.5,
        "cache_dir": "/tmp/dci_",
    },
    "dataset_args": {
        "max_length": 512,
        "mlm_probability": 0.15,
        "plm_probability": 0.1666,
        "max_span_length": 5,
        "batch_size": 8,
        "train_file": "ade_corpus_v2",
        "validation_file": "ade_corpus_v2",
        "cache_dir": "/tmp/dci_",
    },
    "trainer_args": {
        "default_root_dir": "here",
        "val_check_interval": 5000,
        "max_epochs": 1,
        "accumulate_grad_batches": 1,
        "limit_val_batches": 500,
        "log_every_n_steps": 500,
        "monitor": "val_loss",
        "save_top_k": 2,
        "mode": "min",
        "every_n_train_steps": 50000,
    },
}


def test_get_data_and_model_modules_mlm():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    test_pipeline = cast(LanguageModelingTrainingPipeline, pipeline())

    config = template_config.copy()

    file_path = pkg_resources.resource_filename(
        "gt4sd",
        "training_pipelines/tests/lm_example.jsonl",
    )

    config["dataset_args"]["train_file"] = file_path

    config["dataset_args"]["validation_file"] = file_path
    config["model_args"]["type"] = "mlm"

    data_module, model_module = test_pipeline.get_data_and_model_modules(
        config["model_args"], config["dataset_args"]  # type: ignore
    )

    assert isinstance(model_module, MLMModule)
    assert isinstance(data_module, MLMDataModule)

    check_model_config(model_module, config["model_args"])
    check_data_config(data_module, config["dataset_args"])


def test_get_data_and_model_modules_clm():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    test_pipeline = cast(LanguageModelingTrainingPipeline, pipeline())

    config = template_config.copy()

    file_path = pkg_resources.resource_filename(
        "gt4sd",
        "training_pipelines/tests/lm_example.jsonl",
    )

    config["dataset_args"]["train_file"] = file_path
    config["dataset_args"]["validation_file"] = file_path
    config["model_args"]["type"] = "clm"
    config["model_args"]["model_name_or_path"] = "gpt2"

    data_module, model_module = test_pipeline.get_data_and_model_modules(
        config["model_args"], config["dataset_args"]  # type: ignore
    )

    assert isinstance(model_module, CLMModule)
    assert isinstance(data_module, CLMDataModule)

    check_model_config(model_module, config["model_args"])
    check_data_config(data_module, config["dataset_args"])


def test_get_data_and_model_modules_cgm():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    test_pipeline = cast(LanguageModelingTrainingPipeline, pipeline())

    config = template_config.copy()

    file_path = pkg_resources.resource_filename(
        "gt4sd",
        "training_pipelines/tests/lm_example.jsonl",
    )

    config["dataset_args"]["train_file"] = file_path
    config["dataset_args"]["validation_file"] = file_path
    config["model_args"]["type"] = "cgm"
    config["model_args"]["model_name_or_path"] = "t5-base"

    data_module, model_module = test_pipeline.get_data_and_model_modules(
        config["model_args"], config["dataset_args"]  # type: ignore
    )

    assert isinstance(model_module, CGMModule)
    assert isinstance(data_module, CGMDataModule)

    check_model_config(model_module, config["model_args"])
    check_data_config(data_module, config["dataset_args"])


def test_get_data_and_model_modules_plm():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    test_pipeline = cast(LanguageModelingTrainingPipeline, pipeline())

    config = template_config.copy()

    file_path = pkg_resources.resource_filename(
        "gt4sd",
        "training_pipelines/tests/lm_example.jsonl",
    )

    config["dataset_args"]["train_file"] = file_path
    config["dataset_args"]["validation_file"] = file_path
    config["model_args"]["type"] = "plm"
    config["model_args"]["model_name_or_path"] = "xlnet-base-cased"

    data_module, model_module = test_pipeline.get_data_and_model_modules(
        config["model_args"], config["dataset_args"]  # type: ignore
    )

    assert isinstance(model_module, PLMModule)
    assert isinstance(data_module, PLMDataModule)

    check_model_config(model_module, config["model_args"])
    check_data_config(data_module, config["dataset_args"])


def test_add_callbacks():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    callbacks_input = {
        "model_checkpoint_callback": {
            "monitor": "val_loss",
            "save_top_k": 2,
            "mode": "min",
            "every_n_train_steps": 50000,
        }
    }

    callbacks = pipeline().add_callbacks(callbacks_input)  # type: ignore

    assert len(callbacks) == 1
    assert isinstance(callbacks[0], ModelCheckpoint)


def check_model_config(module, config):
    for entry in module.model_args:
        assert entry in config
        assert config[entry] == module.model_args[entry]


def check_data_config(module, config):
    for entry in module.dataset_args:
        assert entry in config
        assert config[entry] == module.dataset_args[entry]



"""Module initialization for gt4sd traning pipelines."""

import json
import logging
from typing import Any, Dict

import pkg_resources

from ..cli.load_arguments_from_dataclass import extract_fields_from_class
from .paccmann.core import PaccMannDataArguments, PaccMannTrainingArguments
from .paccmann.vae.core import PaccMannVAEModelArguments, PaccMannVAETrainingPipeline
from .pytorch_lightning.core import PytorchLightningTrainingArguments
from .pytorch_lightning.granular.core import (
    GranularDataArguments,
    GranularModelArguments,
    GranularTrainingPipeline,
)
from .pytorch_lightning.language_modeling.core import (
    LanguageModelingDataArguments,
    LanguageModelingModelArguments,
    LanguageModelingTrainingPipeline,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

TRAINING_PIPELINE_NAME_METADATA_MAPPING = {
    "mock_training_pipeline": "mock_training_pipeline.json",
    "Terminator training": "terminator_training.json",
}

TRAINING_PIPELINE_ARGUMENTS_MAPPING = {
    "language-modeling-trainer": (
        PytorchLightningTrainingArguments,
        LanguageModelingDataArguments,
        LanguageModelingModelArguments,
    ),
    "paccmann-vae-trainer": (
        PaccMannTrainingArguments,
        PaccMannDataArguments,
        PaccMannVAEModelArguments,
    ),
    "granular-trainer": (
        PytorchLightningTrainingArguments,
        GranularDataArguments,
        GranularModelArguments,
    ),
}

TRAINING_PIPELINE_MAPPING = {
    "language-modeling-trainer": LanguageModelingTrainingPipeline,
    "paccmann-vae-trainer": PaccMannVAETrainingPipeline,
    "granular-trainer": GranularTrainingPipeline,
}


def training_pipeline_name_to_metadata(name: str) -> Dict[str, Any]:
    """Retrieve training pipeline metadata from the name.

    Args:
        name: name of the pipeline.

    Returns:
        dictionary describing the parameters of the pipeline. If the pipeline is not found, no metadata (a.k.a., an empty dictionary is returned).
    """
    metadata: Dict[str, Any] = {"training_pipeline": name, "parameters": {}}
    if name in TRAINING_PIPELINE_NAME_METADATA_MAPPING:
        try:
            with open(
                pkg_resources.resource_filename(
                    "gt4sd",
                    f"training_pipelines/{TRAINING_PIPELINE_NAME_METADATA_MAPPING[name]}",
                ),
                "rt",
            ) as fp:
                metadata["parameters"] = json.load(fp)
        except Exception:
            logger.exception(
                f'training pipeline "{name}" metadata fetching failed, returning an empty metadata dictionary'
            )

    elif name in TRAINING_PIPELINE_ARGUMENTS_MAPPING:

        for training_argument_class in TRAINING_PIPELINE_ARGUMENTS_MAPPING[name]:
            field_types = extract_fields_from_class(training_argument_class)
            metadata["parameters"].update(field_types)

    else:
        logger.warning(
            f'training pipeline "{name}" metadata not found, returning an empty metadata dictionary'
        )
    metadata["description"] = metadata["parameters"].pop(
        "description", "A training pipeline."
    )
    return metadata



"""MolGX Algorithm.

MolGX generation algorithm.
"""

import logging
from dataclasses import field
from typing import Any, ClassVar, Dict, Iterator, Optional, TypeVar

from ....extras import EXTRAS_ENABLED

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

if EXTRAS_ENABLED:
    from ....domains.materials import SmallMolecule, validate_molecules
    from ....exceptions import InvalidItem
    from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
    from ...registry import ApplicationsRegistry
    from .implementation import MolGXGenerator

    T = type(None)
    S = TypeVar("S", bound=SmallMolecule)

    class MolGX(GeneratorAlgorithm[S, T]):
        """MolGX Algorithm."""

        def __init__(
            self,
            configuration: AlgorithmConfiguration[S, T],
            target: Optional[T] = None,
        ):
            """Instantiate MolGX ready to generate items.

            Args:
                configuration: domain and application
                    specification defining parameters, types and validations.
                target: a target for which to generate items.

            Example:
                An example for generating small molecules (SMILES) with given HOMO and LUMO energies:

                    configuration = MolGXQM9Generator()
                    molgx = MolGX(configuration=configuration, target=target)
                    items = list(molgx.sample(10))
                    print(items)
            """

            configuration = self.validate_configuration(configuration)
            # TODO there might also be a validation/check on the target input

            super().__init__(
                configuration=configuration,  # type:ignore
                target=target,  # type:ignore
            )

        def get_generator(
            self,
            configuration: AlgorithmConfiguration[S, T],
            target: Optional[T],
        ) -> Untargeted:
            """Get the function to sample batches via the ConditionalGenerator.

            Args:
                configuration: helps to set up the application.
                target: context or condition for the generation. Unused in the algorithm.

            Returns:
                callable generating a batch of items.
            """
            logger.info("ensure artifacts for the application are present.")
            self.local_artifacts = configuration.ensure_artifacts()
            implementation: MolGXGenerator = configuration.get_conditional_generator(  # type: ignore
                self.local_artifacts
            )
            return implementation.generate

        def validate_configuration(
            self, configuration: AlgorithmConfiguration[S, T]
        ) -> AlgorithmConfiguration[S, T]:
            # TODO raise InvalidAlgorithmConfiguration
            assert isinstance(configuration, AlgorithmConfiguration)
            return configuration

        def sample(self, number_of_items: int = 100) -> Iterator[S]:
            """Generate a number of unique and valid items.

            Args:
                number_of_items: number of items to generate.
                    Defaults to 100.

            Yields:
                the items.
            """
            if hasattr(self.configuration, "maximum_number_of_solutions"):
                maxiumum_number_of_molecules = int(
                    getattr(self.configuration, "maximum_number_of_solutions")
                )
                if number_of_items > maxiumum_number_of_molecules:
                    logger.warning(
                        f"current MolGX configuration can not support generation of {number_of_items} molecules..."
                    )
                    logger.warning(
                        f"to enable generation of {number_of_items} molecules, increase 'maximum_number_of_solutions' (currently set to {maxiumum_number_of_molecules})"
                    )
                    number_of_items = maxiumum_number_of_molecules
                    logger.warning(
                        f"generating at most: {maxiumum_number_of_molecules}..."
                    )
            return super().sample(number_of_items=number_of_items)

    @ApplicationsRegistry.register_algorithm_application(MolGX)
    class MolGXQM9Generator(AlgorithmConfiguration[SmallMolecule, Any]):
        """Configuration to generate compounds with given HOMO and LUMO energies."""

        algorithm_type: ClassVar[str] = "generation"
        domain: ClassVar[str] = "materials"
        algorithm_version: str = "v0"

        homo_energy_value: float = field(
            default=-0.25,
            metadata=dict(description="Target HOMO energy value."),
        )
        lumo_energy_value: float = field(
            default=0.08,
            metadata=dict(description="Target LUMO energy value."),
        )
        use_linear_model: bool = field(
            default=True,
            metadata=dict(description="Linear model usage."),
        )
        number_of_candidates: int = field(
            default=2,
            metadata=dict(description="Number of candidates to consider."),
        )
        maximum_number_of_candidates: int = field(
            default=5,
            metadata=dict(description="Maximum number of candidates to consider."),
        )
        maximum_number_of_solutions: int = field(
            default=10,
            metadata=dict(description="Maximum number of solutions."),
        )
        maximum_number_of_nodes: int = field(
            default=50000,
            metadata=dict(
                description="Maximum number of nodes in the graph exploration."
            ),
        )
        beam_size: int = field(
            default=2000,
            metadata=dict(description="Size of the beam during search."),
        )
        without_estimate: bool = field(
            default=True,
            metadata=dict(description="Disable estimates."),
        )
        use_specific_rings: bool = field(
            default=True,
            metadata=dict(
                description="Flag to indicate whether specific rings are used."
            ),
        )
        use_fragment_const: bool = field(
            default=False,
            metadata=dict(description="Using constant fragments."),
        )

        def get_target_description(self) -> Optional[Dict[str, str]]:
            """Get description of the target for generation.

            Returns:
                target description, returns None in case no target is used.
            """
            return None

        def get_conditional_generator(self, resources_path: str) -> MolGXGenerator:
            """Instantiate the actual generator implementation.

            Args:
                resources_path: local path to model files.

            Returns:
                instance with :meth:`generate<gt4sd.algorithms.conditional_generation.molgx.implementation.MolGXGenerator.generate>` for generation.
            """
            return MolGXGenerator(
                resources_path=resources_path,
                homo_energy_value=self.homo_energy_value,
                lumo_energy_value=self.lumo_energy_value,
                use_linear_model=self.use_linear_model,
                number_of_candidates=self.number_of_candidates,
                maximum_number_of_candidates=self.maximum_number_of_candidates,
                maximum_number_of_solutions=self.maximum_number_of_solutions,
                maximum_number_of_nodes=self.maximum_number_of_nodes,
                beam_size=self.beam_size,
                without_estimate=self.without_estimate,
                use_specific_rings=self.use_specific_rings,
                use_fragment_const=self.use_fragment_const,
                tag_name="qm9",
            )

        def validate_item(self, item: str) -> SmallMolecule:
            """Check that item is a valid SMILES.

            Args:
                item: a generated item that is possibly not valid.

            Raises:
                InvalidItem: in case the item can not be validated.

            Returns:
                the validated SMILES.
            """
            (
                molecules,
                _,
            ) = validate_molecules([item])
            if molecules[0] is None:
                raise InvalidItem(
                    title="InvalidSMILES",
                    detail=f'rdkit.Chem.MolFromSmiles returned None for "{item}"',
                )
            return SmallMolecule(item)


else:
    logger.warning("install AMD_analytcs extras to use MolGX")



"""Dataset module."""

import logging
import os
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

import pandas as pd
import torch
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder, StandardScaler
from torch.utils.data import Dataset

from ..ml.models import ARCHITECTURE_FACTORY, AUTOENCODER_ARCHITECTURES
from ..tokenizer.tokenizer import TOKENIZER_FACTORY, Tokenizer

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

SCALING_FACTORY_FN: Dict[str, Callable] = {
    "onehot": lambda: OneHotEncoder(handle_unknown="error", sparse=False),
    "min-max": lambda: MinMaxScaler(),
    "standard": lambda: StandardScaler(),
}
MODEL_TYPES = set(ARCHITECTURE_FACTORY.keys())


class GranularDataset(Dataset):
    """A dataset wrapper for granular"""

    def __init__(self, name: str, data: Dict[str, Any]) -> None:
        """Initialize a granular dataset.

        Args:
            name: dataset name.
            data: dataset samples.
        """
        self.dataset: Dict[str, Any] = {"name": name, "data": data}

    def __len__(self) -> int:
        """Dataset length.

        Returns:
            length of the dataset.
        """
        lengths = {key: len(data) for key, data in self.dataset["data"].items()}
        if len(set(lengths.values())) > 1:
            raise ValueError(f"mismatching dimensions for the data: {lengths}")
        return list(lengths.values())[0]

    def __getitem__(self, index: int) -> Dict[str, Any]:
        """Retrieve an item from the dataset by index.

        Args:
            index: index for the item.

        Returns:
            an item.
        """
        result = dict()
        for key in self.dataset["data"]:
            result[self.dataset["name"] + "_" + key] = self.dataset["data"][key][index]
        return result


class CombinedGranularDataset(Dataset):
    """General dataset combining multiple granular datasets."""

    def __init__(self, datasets: List[Dict[str, Any]]) -> None:
        """Initialize a general dataset.

        Args:
            datasets: list of dataset configurations.
        """
        self.datasets = datasets
        self.names = [data["name"] for data in datasets]

    def __len__(self) -> int:
        """Dataset length.

        Returns:
            length of the dataset.
        """
        return len([*self.datasets[0]["data"].values()][0])

    def __getitem__(self, index: int) -> Dict[str, Any]:
        """Retrieve an item from the dataset by index.

        Args:
            index: index for the item.

        Returns:
            an item.
        """
        result = dict()
        for dataset in self.datasets:
            keys = [*dataset["data"]]
            for key in keys:
                result[dataset["name"] + "_" + key] = dataset["data"][key][index]
        return result


class SmilesTokenizationPreProcessingDataset(GranularDataset):
    """Dataset for SMILES/SELFIES preprocessing."""

    def __init__(
        self,
        name: str,
        data_columns: Dict[str, Any],
        input_smiles: pd.DataFrame,
        target_smiles: pd.DataFrame,
        tokenizer: Tokenizer,
        set_seq_size: Optional[int] = None,
    ) -> None:
        """Construct a SmilesTokenizationPreProcessingDataset.

        Args:
            name: dataset name.
            data_columns: data columns mapping.
            input_smiles: dataframe containing input SMILES.
            target_smiles: dataframe containing target SMILES.
            tokenizer: a tokenizer defining the molecule representation used.
            set_seq_size: sequence size. Defaults to None, a.k.a., define this
                using the input SMILES.
        """
        self.name = name
        self.input_smiles = input_smiles.values.flatten().tolist()
        self.target_smiles = target_smiles.values.flatten().tolist()
        self.tokenizer = tokenizer
        self.input_tokens: List[torch.Tensor] = []
        self.target_tokens: List[torch.Tensor] = []

        tokens_ids = [
            tokenizer.convert_tokens_to_ids(tokenizer.tokenize(smile))
            for smile in self.input_smiles
        ]
        if set_seq_size:
            self.set_seq_size = set_seq_size
        else:
            self.set_seq_size = max([len(i) for i in tokens_ids]) + 20

        self.smiles_to_ids(input_smiles=self.input_smiles)
        self.smiles_to_ids(target_smiles=self.target_smiles)

        super().__init__(
            name=name,
            data={
                data_columns["input"]: self.input_tokens,
                data_columns["target"]: self.target_tokens,
            },
        )

    def smiles_to_ids(
        self, input_smiles: List[str] = [], target_smiles: List[str] = []
    ) -> None:
        """Process input SMILES lists generating examples by tokenizing strings and converting them to tensors.

        Args:
            input_smiles: list of input SMILES representations. Defaults to [].
            target_smiles: list of target SMILES representations. Defaults to [].
        """
        if len(input_smiles) > 0 and len(target_smiles) == 0:
            self.input_smiles = input_smiles
            smiles = input_smiles
        elif len(input_smiles) == 0 and len(target_smiles) > 0:
            self.target_smiles = target_smiles
            smiles = target_smiles
        else:
            raise Exception(
                "Either input_smiles or target_smiles needs to be specified"
            )

        tokens_ids = [
            self.tokenizer.convert_tokens_to_ids(self.tokenizer.tokenize(smile))
            for smile in smiles
        ]
        examples = []
        for token in tokens_ids:
            example_tokens = self.tokenizer.convert_tokens_to_ids(
                [self.tokenizer.sos_token]
            )
            example_tokens.extend(token)
            example_tokens.extend(
                self.tokenizer.convert_tokens_to_ids([self.tokenizer.eos_token])
            )
            examples.append(
                torch.tensor(
                    self.tokenizer.add_padding_tokens(example_tokens, self.set_seq_size)
                )
            )

        if len(input_smiles) > 0 and len(target_smiles) == 0:
            self.input_tokens = examples
        elif len(input_smiles) == 0 and len(target_smiles) > 0:
            self.target_tokens = examples


class LatentModelDataset(GranularDataset):
    """Latent model dataset."""

    def __init__(
        self,
        name: str,
        data_columns: Dict[str, Any],
        target_data: pd.DataFrame,
        scaling: Optional[str] = None,
    ) -> None:
        """Construct a LatentModelDataset.

        Args:
            name: dataset name.
            data_columns: data columns mapping.
            target_data: dataframe for targets.
            scaling: feature scaling process. Defaults to None, a.k.a. no scaling. Currently not supported.

        Raises:
            NotImplementedError: in case a scaling is selected.
        """
        self.name = name
        if scaling:
            raise NotImplementedError("Scaling not yet supported")
        self.target_data = torch.from_numpy(target_data.values)
        self.target_data = self.target_data.type(torch.float)
        self.target_size = target_data.shape[1]
        super().__init__(name=name, data={data_columns["target"]: self.target_data})


class AutoEncoderDataset(GranularDataset):
    """Autoencoder dataset."""

    def __init__(
        self,
        name: str,
        data_columns: Dict[str, Any],
        input_data: pd.DataFrame,
        target_data: pd.DataFrame,
        scaling: Optional[str] = None,
    ) -> None:
        """Construct an AutoEncoderDataset.

        Args:
            name: dataset name.
            data_columns: data columns mapping.
            input_data: dataframe for inputs.
            target_data: dataframe for targets.
            scaling: feature scaling process. Defaults to None, a.k.a. no scaling. Feasible values: "onehot", "min-max" and "standard".

        Raises:
            ValueError: in case requested scaling is not supported.
        """
        self.name = name
        self.data_columns = data_columns

        if scaling is None:
            self.input_data = torch.from_numpy(input_data.values)
            self.target_data = torch.from_numpy(target_data.values)
        else:
            if scaling not in SCALING_FACTORY_FN:
                raise ValueError(
                    f"Scaling={scaling} not supported. Pick a valid one: {sorted(list(SCALING_FACTORY_FN.keys()))}"
                )

            self.input_scaling = ColumnTransformer(
                transformers=[
                    (
                        "InputScaling",
                        SCALING_FACTORY_FN[scaling](),
                        [data_columns["input"]],
                    )
                ]
            )
            self.target_scaling = ColumnTransformer(
                transformers=[
                    (
                        "TargetScaling",
                        SCALING_FACTORY_FN[scaling](),
                        [data_columns["target"]],
                    )
                ]
            )

            self.input_data = torch.from_numpy(
                self.input_scaling.fit_transform(pd.concat([input_data], axis=1))
            )
            self.target_data = torch.from_numpy(
                self.target_scaling.fit_transform(pd.concat([target_data], axis=1))
            )

        self.input_data, self.target_data = (
            self.input_data.type(torch.float),
            self.target_data.type(torch.float),
        )
        self.input_size = self.input_data.shape[1]
        self.target_size = self.target_data.shape[1]

        super().__init__(
            name=name,
            data={
                data_columns["input"]: self.input_data,
                data_columns["target"]: self.target_data,
            },
        )


DATASET_FACTORY: Dict[str, Type[GranularDataset]] = {
    "latentmodel": LatentModelDataset,
    "smiles": SmilesTokenizationPreProcessingDataset,
    "selfies": SmilesTokenizationPreProcessingDataset,
    "autoencoder": AutoEncoderDataset,
}


def build_data_columns(hparams: Dict[str, Any]) -> Dict[str, Any]:
    """Build data columns from hyper-parameters.

    Args:
        hparams: hyper-parameters for the data columns.

    Returns:
        data columns.
    """
    try:
        input_columns = hparams["input"]
    except KeyError:
        input_columns = None
    try:
        target_columns = hparams["target"]
    except KeyError:
        target_columns = None
    # create dictionary
    if input_columns:
        data_columns = {"input": input_columns, "target": target_columns}
    else:
        data_columns = {"target": target_columns}
    return data_columns


def build_dataset(
    name: str,
    data: pd.DataFrame,
    dataset_type: str,
    data_columns: Dict[str, Any],
    hparams: Dict[str, Any],
) -> GranularDataset:
    """Build a granular dataset.

    Args:
        name: dataset name.
        data: dataframe representing the dataset.
        dataset_type: dataset type. Feasible values: "latentmodel", "smiles", "selfies" and "autoencoder".
        data_columns: data columns mapping.
        hparams: hyper-parameters for the data columns.

    Raises:
            ValueError: in case requested dataset type is not supported.

    Returns:
        a granular dataset.
    """
    dataset: GranularDataset
    dataset_type = dataset_type.lower()
    if dataset_type not in DATASET_FACTORY:
        raise ValueError(
            f"dataset_type={dataset_type} not supported. Pick a valid one: {sorted(list(DATASET_FACTORY.keys()))}"
        )

    input_columns: List[Any]
    if not dataset_type == "latentmodel":
        if data_columns["input"] == "all":
            input_columns = data.columns.tolist()
        else:
            if isinstance(data_columns["input"], list):
                input_columns = data_columns["input"]
            else:
                input_columns = [data_columns["input"]]

    target_columns: List[Any]
    if data_columns["target"] == "all":
        target_columns = data.columns.tolist()
    else:
        if isinstance(data_columns["target"], list):
            target_columns = data_columns["target"]
        else:
            target_columns = [data_columns["target"]]

    if dataset_type in {"smiles", "selfies"}:
        try:
            build_vocab = hparams["build_vocab"]
        except KeyError:
            build_vocab = None
        try:
            sequence_size = hparams["sequence_size"]
        except KeyError:
            sequence_size = None
        vocab_file = hparams["vocab_file"]

        # build tokenizer
        if build_vocab:
            tokenizer = TOKENIZER_FACTORY[dataset_type](
                vocab_file, smiles=data[input_columns].squeeze().tolist()
            )
        else:
            tokenizer = TOKENIZER_FACTORY[dataset_type](vocab_file, smiles=[])
        dataset = SmilesTokenizationPreProcessingDataset(
            name=name,
            data_columns=data_columns,
            input_smiles=data[input_columns],
            target_smiles=data[target_columns],
            tokenizer=tokenizer,
            set_seq_size=sequence_size,
        )
    elif dataset_type == "latentmodel":
        dataset = LatentModelDataset(
            name=name,
            data_columns=data_columns,
            target_data=data[target_columns],
            scaling=None,
        )
    elif dataset_type == "autoencoder":
        dataset = AutoEncoderDataset(
            name=name,
            data_columns=data_columns,
            input_data=data[input_columns],
            target_data=data[target_columns],
            scaling=hparams["scaling"],
        )

    return dataset


def build_architecture(
    model_type: str,
    data_columns: Dict[str, Any],
    dataset: GranularDataset,
    hparams: Dict[str, Any],
) -> Dict[str, Any]:
    """Build architecture configuration for the selected model type and dataset.

    Args:
        model_type: model type. Feasible values: "vae_rnn", "vae_trans", "mlp_predictor", "no_encoding", "mlp_autoencoder" and "vae_mlp".
        data_columns: data columns mapping.
        dataset: a granular dataset.
        hparams: hyper-parameters for the data columns.

    Raises:
        ValueError: in case requested model type is not supported.

    Returns:
        architecture configuration.
    """
    model_type = model_type.lower()
    if model_type not in MODEL_TYPES:
        raise ValueError(
            f"model_type={model_type} not supported. Pick a valid one: {sorted(list(MODEL_TYPES))}"
        )

    architecture: Dict[str, Any] = {
        "name": hparams["name"],
        "type": hparams["type"],
        "start_from_checkpoint": hparams["start_from_checkpoint"],
        "freeze_weights": hparams["freeze_weights"],
        "data": data_columns,
        "hparams": hparams,
    }

    if model_type in AUTOENCODER_ARCHITECTURES:
        architecture["position"] = hparams["position"]
        if model_type in {"vae_rnn", "vae_trans"}:
            hparams["tokenizer"] = dataset.tokenizer
            hparams["vocab_size"] = dataset.tokenizer.vocab_size
            if model_type == "vae_rnn":
                hparams["embedding_size"] = dataset.set_seq_size
            else:  # "vae_trans"
                hparams["sequence_len"] = dataset.set_seq_size
        elif model_type == "no_encoding":
            hparams["latent_size"] = dataset.input_size
        elif model_type in {"mlp_autoencoder", "vae_mlp"}:
            hparams["input_size_enc"] = dataset.input_size
            hparams["output_size_dec"] = dataset.target_size
    else:  # "mlp_predictor"
        hparams["output_size"] = dataset.target_size
        architecture["from_position"] = hparams["from_position"]

    return architecture


def build_dataset_and_architecture(
    name: str,
    data_path: str,
    data_file: str,
    dataset_type: str,
    model_type: str,
    hparams: Dict[str, Any],
    **kwargs,
) -> Tuple[GranularDataset, Dict[str, Any]]:
    """Build a dataset and an architecture configuration.

    Args:
        name: dataset name.
        data_path: path to the dataset.
        data_file: data file name.
        dataset_type: dataset type. Feasible values: "latentmodel", "smiles", "selfies" and "autoencoder".
        model_type: model type. Feasible values: "vae_rnn", "vae_trans", "mlp_predictor", "no_encoding", "mlp_autoencoder" and "vae_mlp".
        hparams: hyper-parameters for the data columns.

    Raises:
        ValueError: in case the data file has an unsupported extension/format.

    Returns:
        a tuple containing a granular dataset and a related architecture configuration.
    """
    if data_file.endswith(".csv"):
        data = pd.read_csv(f"{data_path}{os.path.sep}{data_file}")
    elif data_file.endswith(".bz2") or data_file.endswith(".pkl"):
        data = pd.read_pickle(f"{data_path}{os.path.sep}{data_file}")
    else:
        raise ValueError(
            f"data_file={data_file} extension not supported. Use a compatible extension/format: {['.csv', '.bz2', '.pkl']}"
        )
    data_columns = build_data_columns(hparams)
    dataset = build_dataset(name, data, dataset_type, data_columns, hparams)
    architecture = build_architecture(model_type, data_columns, dataset, hparams)
    return dataset, architecture



"""CLaSS implementation."""
import logging
from typing import List, Optional

import torch

from ....extras import EXTRAS_ENABLED

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

if EXTRAS_ENABLED:
    from cog.sample_pipeline import CogMolFiles, get_new_samples, load_vae, mogQ
    from cog.z_classifier import (
        load_z_ba_regressor_with_tape_embeddings,
        load_zregressor_model,
    )
    from pag.sample_pipeline import PAGFiles, filter_heuristic
    from pag.z_classifier import load_lumo_clf_model

    class UnsupportedTargetError(RuntimeError):
        """Error for target sequence with unknown embedding."""

        def __init__(self, title: str, detail: str) -> None:
            """Initialize UnsupportedTargetError.

            Args:
                title: title of the error.
                detail: description of the error.
            """
            self.type = "UnsupportedTargetError"
            self.title = title
            self.detail = detail
            super().__init__(detail)

    class CogMolGenerator:
        def __init__(
            self,
            protein_sequence: str,
            model_files: CogMolFiles,
            n_samples_per_round: int = 32,
            device: str = "cpu",
            # dropout: float = 0.2,
            # dropout: Dropout is disabled in eval mode. Defaults to 0.2.
            num_proteins_selectivity: int = 10,
            temp: float = 1.0,
            max_len: int = 100,
            bindingdb_date: Optional[str] = None,
        ) -> None:
            """CogMol generator.

            Args:
                protein_sequence: the target sequence for which to generate molecules.
                n_samples_per_round: batch size.
                model_files: dedicated NamedTuple for artifact filepaths.
                device: for example 'cpu'.
                num_proteins_selectivity: number of random samples for measuring
                    selectivity. Defaults to 10.

            Raises:
                RuntimeError: in the case extras are disabled.
            """
            if not EXTRAS_ENABLED:
                raise RuntimeError(
                    "Can't instantiate CogMolGenerator, extras disabled!"
                )

            self.n_samples_per_round = n_samples_per_round
            self.temp = temp
            self.max_len = max_len
            self.num_proteins_selectivity = num_proteins_selectivity
            self.bindingdb_date = bindingdb_date
            self.device = device

            self.model = load_vae(
                model_files.vae_model,
                model_files.vae_config,
                model_files.vae_vocab,
                device,
            )

            self.clf = load_z_ba_regressor_with_tape_embeddings(
                model_path=model_files.ba_model_path,
                device=device,
                dims=[2048, 1],
                dropout=0.2,
            )
            self.reg = load_zregressor_model(
                model_path=model_files.qed_regressor_model_path, device=device
            )

            self.protein_z_map = torch.load(f=model_files.protein_z_map)  # device?
            self.protein_emb = self.get_target_embedding(
                protein_sequence=protein_sequence
            )
            self.protein_sequence = protein_sequence

            # set all models to eval
            self.model.eval()
            self.clf.eval()
            self.reg.eval()

            self.Q_xi_a = mogQ(model_files.mog_model_file, device=device)

            self.Q_xi_a.init_attr_classifiers(
                attr_clfs={
                    "binding": self.clf,
                    "qed": self.reg,
                    "non_binding": self.clf,
                },
                clf_targets={
                    "binding": 1,
                    "qed": 0,
                    "non_binding": 0,
                },
                protein_emb_binding=self.protein_emb,
                protein_embedding_map=self.protein_z_map,
                num_proteins_selectivity=self.num_proteins_selectivity,
            )

        def sample_accepted(self, target: Optional[str] = None) -> List[str]:
            if target is not None and target != self.protein_sequence:
                self.protein_sequence = target
                self.protein_emb = self.get_target_embedding(protein_sequence=target)
                self.Q_xi_a.init_attr_classifiers(
                    attr_clfs={
                        "binding": self.clf,
                        "qed": self.reg,
                        "non_binding": self.clf,
                    },
                    clf_targets={
                        "binding": 1,
                        "qed": 0,
                        "non_binding": 0,
                    },
                    protein_emb_binding=self.protein_emb,
                    protein_embedding_map=self.protein_z_map,
                    num_proteins_selectivity=self.num_proteins_selectivity,
                )

            samples = get_new_samples(
                model=self.model,
                Q=self.Q_xi_a,
                n_samples=self.n_samples_per_round,
                max_len=self.max_len,
                temp=self.temp,
            )
            return samples[samples["accept_z"] == 1]["smiles"].tolist()

        def get_target_embedding(self, protein_sequence: str) -> torch.Tensor:
            """Retrieve embedding of target or raise a dedicated exception.

            Args:
                protein_sequence: target amino acid sequence.

            Raises:
                UnsupportedTargetError: in case the embedding is not available in
                    `self.protein_z_map`.

            Returns:
                The protein embedding.
            """
            try:
                return self.protein_z_map[protein_sequence]
            except KeyError:
                detail = (
                    "The provided target is not available in this version: \n"
                    f"{protein_sequence}"
                )
                # prepend a hint on the supported target proteins
                if self.bindingdb_date is not None:
                    detail = (
                        f"Only protein sequences published in BindingDB until {self.bindingdb_date} are supported. "
                    ) + detail

                logger.warning(detail)
                raise UnsupportedTargetError(
                    title="The target protein sequence is not supported.", detail=detail
                )

    class PAGGenerator:
        def __init__(
            self,
            model_files: PAGFiles,
            n_samples_per_round: int = 32,
            device: str = "cpu",
            temp: float = 1.0,
            max_len: int = 100,
        ) -> None:
            """PAG generator.

            Args:
                n_samples_per_round: batch size.
                model_files: dedicated NamedTuple for artifact filepaths.
                device: for example 'cpu'.

            Raises:
                RuntimeError: in the case extras are disabled.
            """
            if not EXTRAS_ENABLED:
                raise RuntimeError("Can't instantiate PAGGenerator, extras disabled!")

            self.n_samples_per_round = n_samples_per_round
            self.temp = temp
            self.max_len = max_len
            self.device = device

            self.model = load_vae(
                model_files.vae_model,
                model_files.vae_config,
                model_files.vae_vocab,
                device,
            )

            self.clf = load_lumo_clf_model(model_path=model_files.lumo_clf_model_path)

            # set all models to eval
            self.model.eval()
            self.clf.eval()

            self.Q_xi_a = mogQ(model_files.mog_model_file, device=device)

            self.Q_xi_a.init_attr_classifiers(
                attr_clfs={
                    "LUMO": self.clf,
                },
                clf_targets={
                    "LUMO": 1,
                },
            )

        def sample_accepted(self) -> List[str]:
            samples = get_new_samples(
                model=self.model,
                Q=self.Q_xi_a,
                n_samples=self.n_samples_per_round,
                max_len=self.max_len,
                temp=self.temp,
            )
            # TODO: allow user input to heuristics?
            samples = filter_heuristic(samples)
            return samples[samples.accept_z & samples.accept].smiles.tolist()


else:
    logger.warning("install cogmol-inference extras to use CLaSS")



import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, TypeVar

from ...core import AlgorithmConfiguration, GeneratorAlgorithm
from ...registry import ApplicationsRegistry
from .implementation import ReinventConditionalGenerator

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=Any)
Targeted = Callable[[T], Iterable[Any]]

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Reinvent(GeneratorAlgorithm[S, T]):
    """Reinvent sample generation algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """Instantiate Reinvent ready to generate samples.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for predicting topics for a given text::

                config = ReinventGenerator()
                algorithm = Reinvent(configuration=config, target="")
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to perform the prediction via Reinvent's generator.

        Args:
            configuration: helps to set up specific application of Reinvent.
            target: context or condition for the generation.

        Returns:
            callable with target generating samples.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: ReinventConditionalGenerator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate_samples


@ApplicationsRegistry.register_algorithm_application(Reinvent)
class ReinventGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate molecules using the REINVENT algorithm. It generates the molecules minimizing the distances between the scaffolds."""

    algorithm_name: ClassVar[str] = Reinvent.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=20,
        metadata=dict(description=("Number of samples to generate per scaffold")),
    )

    randomize: bool = field(
        default=True,
        metadata=dict(description=("Randomize the scaffolds if set to true")),
    )

    sample_uniquely: bool = field(
        default=True,
        metadata=dict(description=("Generate unique sample sequences if set to true")),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "SMILES for sample generation",
            "description": "SMILES considered for the samples generation.",
            "type": "string",
        }

    def get_conditional_generator(
        self, resources_path: str
    ) -> ReinventConditionalGenerator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_samples<gt4sd.algorithms.conditional_generation.reinvent.implementation.ReinventConditionalGenerator.generate_samples>` method for targeted generation.
        """
        return ReinventConditionalGenerator(
            resources_path=resources_path,
            batch_size=self.batch_size,
            randomize=self.randomize,
            sample_uniquely=self.sample_uniquely,
        )



#!/usr/bin/env python

"""Transformers pretrained model to SentenceTransformer model converter."""


import json
import logging
import os
import sys
from dataclasses import dataclass, field
from typing import cast

from sentence_transformers import SentenceTransformer, __version__, models

from .argument_parser import ArgumentParser, DataClassType


@dataclass
class TransformersToSentenceTransformersArguments:
    """Transformers to Sentence Transformers converter arguments."""

    __name__ = "hf_to_st_converter_args"

    model_name_or_path: str = field(
        metadata={"help": "HF model name or path."},
    )
    pooling: str = field(
        metadata={
            "help": "Comma separated pooling modes. Supported types: cls, max, mean, mean_sqrt."
        },
    )
    output_path: str = field(
        metadata={"help": "Path to the converted model."},
    )


def main() -> None:
    """Convert HF pretrained model to SentenceTransformer.

    Create a SentenceTransformer model having a given HF model as
    word embedding model plus an optional pooling layer. We can
    also concatenate multiple poolings together.

    Parsing from the command line the following parameters:
        - HF pretrained model to be used as word embedding model.
        - the pooling mode (more than one can be provided as a list), the implemented
            options are "cls", "max", "mean", "mean" and "sqrt".
        - path to save the generated SentenceTransformer model.
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    arguments = ArgumentParser(
        cast(DataClassType, TransformersToSentenceTransformersArguments)
    ).parse_args_into_dataclasses(return_remaining_strings=True)[0]

    model_name_or_path = arguments.model_name_or_path
    pooling = [
        polling_argument.strip() for polling_argument in arguments.pooling.split(",")
    ]
    output_path = arguments.output_path

    word_embedding_model = models.Transformer(model_name_or_path)

    pooling_mode_cls_token = False
    pooling_mode_max_tokens = False
    pooling_mode_mean_tokens = False
    pooling_mode_mean_sqrt_len_tokens = False

    if "cls" in pooling:
        pooling_mode_cls_token = True
    if "max" in pooling:
        pooling_mode_max_tokens = True
    if "mean" in pooling:
        pooling_mode_mean_tokens = True
    if "mean_sqrt" in pooling:
        pooling_mode_mean_sqrt_len_tokens = True

    pooling_model = models.Pooling(
        word_embedding_model.get_word_embedding_dimension(),
        pooling_mode=None,
        pooling_mode_cls_token=pooling_mode_cls_token,
        pooling_mode_max_tokens=pooling_mode_max_tokens,
        pooling_mode_mean_tokens=pooling_mode_mean_tokens,
        pooling_mode_mean_sqrt_len_tokens=pooling_mode_mean_sqrt_len_tokens,
    )

    model = SentenceTransformer(modules=[word_embedding_model, pooling_model])

    model.save(output_path)

    config_filepath = os.path.join(output_path, "config.json")
    if os.path.exists(config_filepath):
        with open(config_filepath) as fp:
            config = json.load(fp)
            config["__version__"] = __version__
        with open(config_filepath, "wt") as fp:
            json.dump(config, fp, indent=2)


if __name__ == "__main__":
    main()



"""Generic utils for pytorch."""

from typing import Optional, Union

import torch


def get_device() -> torch.device:
    """
    Get device dynamically.
    """
    return torch.device("cuda" if torch.cuda.is_available() else "cpu")


def device_claim(device: Optional[Union[torch.device, str]] = None) -> torch.device:
    """
    Satidfy a device claim.

    Args:
        device: device where the inference
            is running either as a dedicated class or a string. If not provided is inferred.

    Returns:
        torch.device: the claimed device or a default one.
    """
    if isinstance(device, str):
        device = torch.device(device)
    device = (
        get_device()
        if (device is None or not isinstance(device, torch.device))
        else device
    )
    return device


def get_device_from_tensor(tensor: torch.Tensor) -> torch.device:
    """Get the device from a tensor.

    Args:
        tensor: a tensor.

    Returns:
        the device.
    """
    device_id = tensor.get_device()
    device = "cpu" if device_id < 0 else f"cuda:{device_id}"
    return device_claim(device)



"""Implementation details for a Template algorithm"""

import random
from typing import List


class Generator:
    """Basic Generator for the template algorithm"""

    def __init__(self, resources_path: str, temperature: int) -> None:
        """Initialize the Generator.

        Args:
            resources_path: directory where to find models and parameters.

        """

        self.resources_path = resources_path
        self.temperature = temperature

    def hello_name(
        self,
        name: str,
    ) -> List[str]:
        """Validate a list of strings.

        Args:
            name: a string.

        Returns:
            a list containing salutation and temperature converted to fahrenheit.
        """
        return [
            f"Hello {str(name)} {random.randint(1, int(1e6))} times and, fun fact, {str(self.temperature)} celsius equals to {(self.temperature * (9/5) + 32)} fahrenheit."
        ]
